# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import

import os
import inspect
import functools
import warnings
import socket
import datetime
import pprint
import copy
import numpy as np
import quaternion
import scipy.constants as spc
from scipy.interpolate import splev, splrep
from quaternion.numba_wrapper import njit, xrange, GOT_NUMBA
from . import *

if GOT_NUMBA:
    @njit('void(c16[:,:], f8[:])')
    def complex_array_norm(c, s):
        for i in xrange(len(s)):
            s[i] = 0.0
            for j in xrange(c.shape[1]):
                s[i] += c[i, j].real ** 2 + c[i, j].imag ** 2
        return

    @njit('void(c16[:,:], f8[:])')
    def complex_array_abs(c, s):
        for i in xrange(len(s)):
            s[i] = 0.0
            for j in xrange(c.shape[1]):
                s[i] += c[i, j].real ** 2 + c[i, j].imag ** 2
            s[i] = np.sqrt(s[i])
        return
else:
    # If njit above is just the dumb placeholder defined in .numba, that
    # function would be ~1000 times slower (because it would use standard
    # python).  We can do much better with numpy, though this is still ~6
    # times slower than numba for typical arrays.
    def complex_array_norm(c, s):
        s[:] = np.sum(c.real ** 2 + c.imag ** 2, axis=-1)
        return

    def complex_array_abs(c, s):
        s[:] = np.sqrt(np.sum(c.real ** 2 + c.imag ** 2, axis=1))
        return


def waveform_alterations(func):
    """Temporarily increment history depth safely

    This decorator stores the value of `self.__history_depth__`, then increments it by 1, calls the function,
    returns the history depth to its original value, and then returns the result of the function.  This should be
    used on any member function that could alter the waveform on which it is called, or which could return a new
    altered version of the original.

    Typically, within the function itself, you will want to decrement the depth manually just before appending to the
    history -- which will presumably take place at the end of the function.  You do not need to undo this,
    as the decorator will take care of that part.

    """
    @functools.wraps(func)
    def func_wrapper(self, *args, **kwargs):
        if self.__history_depth__ == 0:
            self._append_history('')
        stored_history_depth = self.__history_depth__
        self.__history_depth__ += 1
        result = func(self, *args, **kwargs)
        self.__history_depth__ = stored_history_depth
        return result
    return func_wrapper


def test_without_assertions(errs, val, msg=''):
    """Replacement for np.testing.assert_

    This function should be able to replace `assert_`, but rather than raising an exception, this just adds a
    description of the problem to the `errors` variable.

    """
    if not val:
        try:
            smsg = msg()
        except TypeError:
            smsg = msg
        errs += [smsg]


def test_with_assertions(errs, val, msg=''):
    np.testing.assert_(val, 'Failed assertion:\n\t' + msg)


class _object(object):
    """Useless class to allow multiple inheritance"""

    def __init__(self, *args, **kwargs):
        super(_object, self).__init__()


class WaveformBase(_object):
    """Object containing time, frame, and data, along with related information

    This object is just the base object from which these other classes are derived:
      * WaveformModes
      * WaveformGrid
      * WaveformInDetector
      * WaveformInDetectorFT
    For more specific information, see the documentation of those classes.

    Attributes
    ----------
    t : float array
        Time steps corresponding to other data
    frame : quaternion array
        Rotors taking static basis onto decomposition basis
    data : 2-d array of complex or real numbers
        The nature of this data depends on the derived type.  First index is time, second index depends on type.
    history : list of strings
        As far as possible, all functions applied to the object are recorded in the `history` variable.  In fact,
        the object should almost be able to be recreated using the commands in the history list. Commands taking
        large arrays, however, are shortened -- so the data will not be entirely reconstructable.
    frameType : int
        Index corresponding to `scri.FrameType` appropriate for `data`.
    dataType : int
        Index corresponding to `scri.DataType` appropriate for `data`.
    r_is_scaled_out : bool
        True if the `data` have been multiplied by the appropriate power of radius so that the asymptotic value can
        be finite and nonzero.
    m_is_scaled_out : bool
        True if the `data` have been scaled by the appropriate value of the total mass so that they are dimensionless.
    num : int (read only)
        Automatically assigned number of this object.  The constructor of this type keeps count of the number of
        objects it has created, to assign each object a more-or-less unique ID for use in the history strings.  This
        counter is reset at the beginning of each python session.  Subclasses should automatically have a different
        counter.

    Indexing
    --------
    WaveformBase objects can be indexed much like a numpy array, where the first dimension gives the time indices,
    and the second gives the data-set indices. This will return another WaveformBase object containing slices of the
    original data.

    It is important to note, however, that as with numpy array slices, slicing a WaveformBase will not typically copy
    the original data; the result will simply be a view into the data.  This means that changing the data in the
    slice can change the data in the original.  If you want to make a copy, you should probably use the copy
    constructor: `W2 = WaveformBase(W1)`. It is also possible to use the standard copy.deepcopy method.

    Also note that the first slice dimension corresponds to the indices of the time, but the second dimension may NOT
    correspond to indices for derived types.  In particular, for `WaveformModes`, the second index corresponds to
    modes, because this type enforces completeness of each ell mode.  For the `WaveformBase` type, however,
    the second index does correspond to the second dimension of the data.

    For example,

    >>> W  = WaveformBase()
    >>> W[10:-20]

    will give all columns in the data, but only at times starting with the
    10th time step, and ending one before the -20th time step.  Meanwhile,

    >>> W[10:-20,2]

    will give the same range of times, but only the second column (unless the subclass overrides this behavior,
    as in `WaveformModes`).  Similarly,

    >>> W[10:-20,2:5]

    will return the same range of times, along with the 2,3,4 columns. Note the lack of 5 column, for consistency
    with python's usual slice syntax.

    >>> W[:,:0]

    will return all time steps, along with all `frame` data, but `data` will be empty (because the `:0` term selects
    everything before the 0th element).  Similarly,

    >>> W[:0,:0]

    is empty of all numerical data.

    """

    __num = 0  # Used to count number of Waveforms created

    def __init__(self, *args, **kwargs):
        """Initializer for WaveformBase object

        WaveformBase objects may be created in two ways.  First, by copying an existing WaveformBase object -- in
        which case the only parameter should be that object.  Second, by passing any of the (writable) attributes as
        keywords.

        In both cases, the last step in initialization is to check the validity of the result.  By default,
        this will raise an exception if the result is not valid.  An additional keyword parameter
        `override_exception_from_invalidity` may be set if this is not desired.  This may be necessary if only some
        of the data can be passed in to the initializer, for example.

        Keyword parameters
        ------------------
        t: float array, empty default
        frame : quaternion array, empty default
        data : 2-d complex array, empty default
        history : list of strings, empty default
            This is the list of strings prepended to the history, an additional line is appended, showing the call to
            this initializer.
        frameType : int, defaults to 0 (UnknownFrameType)
            See scri.FrameNames for possible values
        dataType : int, defaults to 0 (UnknownDataType)
            See scri.DataNames for possible values
        r_is_scaled_out : bool, defaults to False
            Set to True if the data represented could approach a nonzero value at Scri
        m_is_scaled_out : bool, defaults to False
            Set to True if the data represented are dimensionless and in units where the total mass is 1
        override_exception_from_invalidity: bool, defaults to False
            If True, report any errors, but do not raise them.
        constructor_statement : str, optional
            If this is present, it will replace the default constructor statement added to the history.  It is
            prepended with a string of the form `'{0} = '.format(self)`, which prints the ID of the resulting object
            (unique to this session only).

        """
        original_kwargs = kwargs.copy()
        super(WaveformBase, self).__init__(*args, **kwargs)  # to ensure proper calling in multiple inheritance
        override_exception_from_invalidity = kwargs.pop('override_exception_from_invalidity', False)
        self.__num = type(self).__num
        self.__history_depth__ = 0
        type(self).__num += 1  # Increment class's instance tracker
        if len(args) == 0:
            self.t = kwargs.pop('t', np.empty((0,), dtype=float))
            self.frame = kwargs.pop('frame', np.empty((0,), dtype=np.quaternion))
            self.data = kwargs.pop('data', np.empty((0, 0), dtype=complex))
            # Information about this object
            self.history = kwargs.pop('history', [])
            self.frameType = kwargs.pop('frameType', UnknownFrameType)
            self.dataType = kwargs.pop('dataType', UnknownDataType)
            self.r_is_scaled_out = kwargs.pop('r_is_scaled_out', False)
            self.m_is_scaled_out = kwargs.pop('m_is_scaled_out', False)
            if 'constructor_statement' in kwargs:
                self._append_history('{0} = {1}'.format(self, kwargs.pop('constructor_statement')))
            else:
                opts = np.get_printoptions()
                np.set_printoptions(threshold=6)
                self._append_history('{0} = {1}(**{2})'.format(self, type(self).__name__,
                                                               pprint.pformat(original_kwargs, indent=4)))
                np.set_printoptions(**opts)
        elif len(args) == 1 and isinstance(args[0], type(self)):
            other = args[0]
            self.t = np.copy(other.t)
            self.frame = np.copy(other.frame)
            self.data = np.copy(other.data)
            # Information about this object
            self.history = other.history[:]
            self.frameType = other.frameType
            self.dataType = other.dataType
            self.r_is_scaled_out = other.r_is_scaled_out
            self.m_is_scaled_out = other.m_is_scaled_out
            self._append_history(['', '{0} = {1}({2})'.format(self, type(self).__name__, other)])
        else:
            raise ValueError("Did not understand input arguments to `{0}` constructor.\n".format(type(self).__name__) +
                             "Note that explicit data values must be passed as keywords,\n" +
                             "whereas objects to be copied must be passed as the sole argument.")
        hostname = socket.gethostname()
        cwd = os.getcwd()
        time = datetime.datetime.now().isoformat()
        self.__history_depth__ = 1
        self.ensure_validity(alter=True, assertions=(not override_exception_from_invalidity))
        self.__history_depth__ = 0
        self._append_history(['hostname = {0}'.format(hostname),
                              'cwd = {0}'.format(cwd),
                              'datetime = {0}'.format(time),
                              version_info()], 1)
        if kwargs:
            warning = '\nIn `{0}` initializer, unused keyword arguments:\n'.format(type(self).__name__)
            warning += pprint.pformat(kwargs, indent=4)
            warnings.warn(warning)

    @waveform_alterations
    def ensure_validity(self, alter=True, assertions=False):
        """Try to ensure that the `WaveformBase` object is valid

        This tests various qualities of the WaveformBase's members that are frequently assumed throughout the code.
        If the optional argument `alter` is `True` (which is the default), this function tries to alter the
        WaveformBase in place to ensure validity.  Note that this is not always possible.  If that is the case,
        an exception may be raised.  For example, if the `t` member is not a one-dimensional array of floats,
        it is not clear what that data should be. Similarly, if the `t` and `data` members have mismatched
        dimensions, there is no way to resolve that automatically.

        Also note that this is almost certainly not be an exhaustive test of all assumptions made in the code.

        If the optional `assertions` argument is `True` (default is `False`), the first test that fails will raise an
        assertion error.

        """
        import numbers

        errors = []
        alterations = []

        if assertions:
            test = test_with_assertions
        else:
            test = test_without_assertions

        # Ensure that the various data are correct and compatible
        test(errors,
             isinstance(self.t, np.ndarray),
             'isinstance(self.t, np.ndarray) # type(self.t)={0}'.format(type(self.t)))
        test(errors,
             self.t.dtype == np.dtype(np.float),
             'self.t.dtype == np.dtype(np.float) # self.t.dtype={0}'.format(self.t.dtype))
        if alter and self.t.ndim == 2 and self.t.shape[1] == 1:
            self.t = self.t[:, 0]
            alterations += ['{0}.t = {0}.t[:,0]'.format(self)]
        test(errors,
             not self.t.size or self.t.ndim == 1,
             'not self.t.size or self.t.ndim==1 # self.t.size={0}; self.t.ndim={1}'.format(self.t.size, self.t.ndim))
        test(errors,
             self.t.size <= 1 or np.all(np.diff(self.t) > 0.0),
             'self.t.size<=1 or np.all(np.diff(self.t)>0.0) '
             '# self.t.size={0}; max(np.diff(self.t))={1}'.format(self.t.size, (
                 max(np.diff(self.t)) if self.t.size > 1 else np.nan)))
        test(errors,
             np.all(np.isfinite(self.t)),
             'np.all(np.isfinite(self.t))')

        if alter and self.frame is None:
            self.frame = np.empty((0,), dtype=np.quaternion)
            alterations += ['{0}.frame = np.empty((0,), dtype=np.quaternion)'.format(self)]
        test(errors,
             isinstance(self.frame, np.ndarray),
             'isinstance(self.frame, np.ndarray) # type(self.frame)={0}'.format(type(self.frame)))
        if alter and self.frame.dtype == np.dtype(np.float):
            try:  # Might fail because of shape
                self.frame = quaternion.as_quat_array(self.frame)
                alterations += ['{0}.frame = quaternion.as_quat_array({0}.frame)'.format(self)]
            except (AssertionError, ValueError):
                pass
        test(errors,
             self.frame.dtype == np.dtype(np.quaternion),
             'self.frame.dtype == np.dtype(np.quaternion) # self.frame.dtype={0}'.format(self.frame.dtype))
        test(errors,
             self.frame.size <= 1 or self.frame.size == self.t.size,
             'self.frame.size<=1 or self.frame.size==self.t.size '
             '# self.frame.size={0}; self.t.size={1}'.format(self.frame.size, self.t.size))
        test(errors,
             np.all(np.isfinite(self.frame)),
             'np.all(np.isfinite(self.frame))')

        test(errors,
             isinstance(self.data, np.ndarray),
             'isinstance(self.data, np.ndarray) # type(self.data)={0}'.format(type(self.data)))
        test(errors,
             self.data.ndim >= 1,
             'self.data.ndim >= 1 # self.data.ndim={0}'.format(self.data.ndim))
        test(errors,
             self.data.shape[0] == self.t.shape[0],
             'self.data.shape[0]==self.t.shape[0] '
             '# self.data.shape[0]={0}; self.t.shape[0]={1}'.format(self.data.shape[0], self.t.shape[0]))
        test(errors,
             np.all(np.isfinite(self.data)),
             'np.all(np.isfinite(self.data))')

        # Information about this object
        if alter and not self.history:
            self.history = ['']
            alterations += ["{0}.history = ['']".format(self)]
        if alter and isinstance(self.history, str):
            self.history = self.history.split('\n')
            alterations += ["{0}.history = {0}.history.split('\n')".format(self)]
        test(errors,
             isinstance(self.history, list),
             'isinstance(self.history, list) # type(self.history)={0}'.format(type(self.history)))
        test(errors,
             isinstance(self.history[0], str),
             'isinstance(self.history[0], str) # type(self.history[0])={0}'.format(type(self.history[0])))
        test(errors,
             isinstance(self.frameType, numbers.Integral),
             'isinstance(self.frameType, numbers.Integral) # type(self.frameType)={0}'.format(type(self.frameType)))
        test(errors,
             self.frameType in FrameType,
             'self.frameType in FrameType # self.frameType={0}'.format(self.frameType))
        test(errors,
             isinstance(self.dataType, numbers.Integral),
             'isinstance(self.dataType, numbers.Integral) # type(self.dataType)={0}'.format(type(self.dataType)))
        test(errors,
             self.dataType in DataType,
             'self.dataType in DataType # self.dataType={0}'.format(self.dataType))
        test(errors,
             isinstance(self.r_is_scaled_out, bool),
             'isinstance(self.r_is_scaled_out, bool) # type(self.r_is_scaled_out)={0}'.format(
                 type(self.r_is_scaled_out)))
        test(errors,
             isinstance(self.m_is_scaled_out, bool),
             'isinstance(self.m_is_scaled_out, bool) # type(self.m_is_scaled_out)={0}'.format(
                 type(self.m_is_scaled_out)))
        test(errors,
             isinstance(self.num, numbers.Integral),
             'isinstance(self.num, numbers.Integral) # type(self.num)={0}'.format(type(self.num)))

        if alterations:
            self._append_history(alterations)
            warnings.warn("The following alterations were made:\n\t" + '\n\t'.join(alterations))
        if errors:
            warnings.warn("The following conditions were found to be incorrectly False:\n\t" + '\n\t'.join(errors))
            return False

        self.__history_depth__ -= 1
        self._append_history('WaveformBase.ensure_validity' +
                             '({0}, alter={1}, assertions={2})'.format(self, alter, assertions))

        return True

    @property
    def is_valid(self):
        return self.ensure_validity(alter=False, assertions=False)

    # Data sizes
    @property
    def n_data_sets(self):
        return int(np.prod(self.data.shape[1:]))

    @property
    def n_times(self):
        return self.t.shape[0]

    # Calculate weights
    @property
    def spin_weight(self):
        return SpinWeights[self.dataType]

    @property
    def conformal_weight(self):
        return (ConformalWeights[self.dataType]
                + (-RScaling[self.dataType] if self.r_is_scaled_out else 0))

    @property
    def gamma_weight(self):
        """Non-conformal effect of a boost.

        This factor allows for mass-scaling, for example.  If the waveform describes `r*h/M`, for example,
        then `r` and `h` vary by the conformal weight, which depends on the direction; whereas `M` is a monopole,
        and thus cannot depend on the direction.  Instead, `M` simply obeys the standard formula, scaling with gamma.

        """
        return ((MScaling[self.dataType] if self.m_is_scaled_out else 0)
                + (-RScaling[self.dataType] if (self.r_is_scaled_out and self.m_is_scaled_out) else 0))

    @property
    def r_scaling(self):
        return RScaling[self.dataType]

    @property
    def m_scaling(self):
        return MScaling[self.dataType]

    # Text descriptions
    @property
    def num(self):
        return self.__num

    @property
    def frame_type_string(self):
        return FrameNames[self.frameType]

    @property
    def data_type_string(self):
        return DataNames[self.dataType]

    @property
    def data_type_latex(self):
        return DataNamesLaTeX[self.dataType]

    @property
    def descriptor_string(self):
        """Create a simple string describing the content of the waveform

        This string will be suitable for file names.  For example, 'rMpsi4' or 'rhOverM'.  It uses the waveform's
        knowledge of itself, so if this is incorrect, the result will be incorrect.

        """
        if self.dataType == UnknownDataType:
            return self.data_type_string
        descriptor = ""
        if self.r_is_scaled_out:
            if RScaling[self.dataType] == 1:
                descriptor = "r"
            elif RScaling[self.dataType] > 1:
                descriptor = "r" + str(RScaling[self.dataType])
        if self.m_is_scaled_out:
            Mexponent = MScaling[self.dataType] - (RScaling[self.dataType] if self.r_is_scaled_out else 0)
            if Mexponent < -1:
                descriptor = descriptor + self.data_type_string + "OverM" + str(-Mexponent)
            elif Mexponent == -1:
                descriptor = descriptor + self.data_type_string + "OverM"
            elif Mexponent == 1:
                descriptor = descriptor + "M" + self.data_type_string
            elif Mexponent > 1:
                descriptor = descriptor + "M" + str(Mexponent) + self.data_type_string
        else:
            descriptor = descriptor + self.data_type_string
        return descriptor

    # Data simplifications
    @property
    def data_2d(self):
        return self.data.reshape((self.n_times, self.n_data_sets))

    @property
    def abs(self):
        return np.abs(self.data)

    @property
    def arg(self):
        return np.angle(self.data)

    @property
    def arg_unwrapped(self):
        return np.unwrap(np.angle(self.data), axis=0)

    def norm(self, take_sqrt=False, indices=slice(None, None, None)):
        """L2 norm of the waveform

        The optional arguments say whether to take the square-root of
        the norm at each time, and allow restriction to a slice of the
        data, respectively.

        """
        if indices == slice(None, None, None):
            n = np.empty((self.n_times,), dtype=float)
        else:
            n = np.empty((self.t[indices].shape[0],), dtype=float)
        if take_sqrt:
            complex_array_abs(self.data_2d[indices], n)
        else:
            complex_array_norm(self.data_2d[indices], n)
        return n

    def max_norm_index(self, skip_fraction_of_data=4):
        """Index of time step with largest norm

        The optional argument skips a fraction of the data.  The default is
        4, which means that it only searches the last three-fourths of the
        data for the max.  If 0 or 1 is input, this is ignored, and all the
        data is searched.

        """
        if skip_fraction_of_data == 0 or skip_fraction_of_data == 1:
            indices = slice(None, None, None)
            return np.argmax(self.norm(indices=indices))
        else:
            indices = slice(self.n_times // skip_fraction_of_data, None, None)
            return np.argmax(self.norm(indices=indices)) + (self.n_times // skip_fraction_of_data)

    def max_norm_time(self, skip_fraction_of_data=4):
        """Return time at which largest norm occurs in data

        See `help(max_norm_index)` for explanation of the optional argument.

        """
        return self.t[self.max_norm_index(skip_fraction_of_data=skip_fraction_of_data)]

    @property
    def data_dot(self):
        return quaternion.calculus.derivative(self.data, self.t)

    # Data representations
    def _append_history(self, hist, additional_depth=0):
        """Add to the object's history log

        Input may be a single string or list of strings.  Any newlines will be split into separate strings.  Each
        such string is then prepended with a number of `#`s, indicating that the content of that line was called from
        within a member function, or is simply a piece of information relevant to the waveform.  The idea behind this
        is that the history should be -- as nearly as possible --  a script that could be run to reproduce the
        waveform, so the lines beginning with `#` would not be run.

        The number of `#`s is controlled by the object's `__history_depth__` field and the optional input to this
        function; their sum is the number prepended.  The user should never have to deal with this issue,
        but all member functions should increment the `__history_depth__` before calling another member function,
        and decrement it again as necessary before recording itself in the history.  Also, for any lines added just
        for informational purposes (e.g., the hostname, pwd, date, and versions added in `__init__`), this function
        should be called with `1` as the optional argument.

        """
        if not isinstance(hist, list):
            hist = [hist]
        self.history += ['# ' * (self.__history_depth__ + additional_depth) + hist_line
                         for hist_element in hist
                         for hist_line in hist_element.split('\n')]

    def __str__(self):
        # "The goal of __str__ is to be readable; the goal of __repr__ is to be unambiguous." --- stackoverflow
        return '{0}_{1}'.format(type(self).__name__, self.num)

    def __repr__(self):
        # "The goal of __str__ is to be readable; the goal of __repr__ is to be unambiguous." --- stackoverflow
        from textwrap import dedent

        opts = np.get_printoptions()
        np.set_printoptions(threshold=6, linewidth=150, precision=6)
        rep = """
         {0}(
             t={1},
             frame={2},
             data={4},
             frameType={5}, dataType={6},
             r_is_scaled_out={7}, m_is_scaled_out={8})  # num = {9}"""
        rep = rep.format(type(self).__name__,
                         str(self.t).replace('\n', '\n' + ' ' * 15),
                         str(self.frame).replace('\n', '\n' + ' ' * 19),
                         self.history, str(self.data).replace('\n', '\n' + ' ' * 18),
                         self.frameType, self.dataType, self.r_is_scaled_out, self.m_is_scaled_out, self.num)
        np.set_printoptions(**opts)
        return dedent(rep)

    def __getstate__(self):
        """Get state of object for copying and pickling

        The only nontrivial operation is with quaternions, since they can't
        currently be pickled automatically.  We just view the frame array as
        a float array, and pickle as usual.

        Also, we remove the `num` value, because this will get reset
        properly on creation.

        """
        state = copy.deepcopy(self.__dict__)
        state['frame'] = quaternion.as_float_array(self.frame)
        return state

    def __setstate__(self, state):
        """Set state of object for copying and pickling

        The only nontrivial operation is with quaternions, since they can't
        currently be pickled automatically.  We just view the frame array as
        a float array, and unpickle as usual, then convert the float array
        back to a quaternion array.

        """
        old_num = state.pop('_WaveformBase__num')
        self.__dict__.update(state)
        self.frame = quaternion.as_quat_array(self.frame)
        self._append_history('copied, deepcopied, or unpickled as {0}'.format(self), 1)
        self._append_history('{0} = {1}'.format(self, '{0}'.format(self).replace(str(self.num), str(old_num))))

    @waveform_alterations
    def deepcopy(self):
        """Return a deep copy of the object

        This is just an alias for `copy`, which is deep anyway.

        """
        W = self.copy()
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.deepcopy()'.format(W, self))
        return W

    @waveform_alterations
    def copy(self):
        """Return a (deep) copy of the object

        Note that this also copies all members if the object is a subclass.  If you want a forgetful WaveformBase
        object, you can simply use the copy constructor.

        """
        W = type(self)()
        state = copy.deepcopy(self.__dict__)
        state.pop('_WaveformBase__num')
        W.__dict__.update(state)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.copy()'.format(W, self))
        return W

    @waveform_alterations
    def copy_without_data(self):
        """Return a copy of the object, with empty `t`, `frame`, and `data` fields

        Note that subclasses may override this to set some related data members.  For example,
        `WaveformModes.copy_without_data` sets the `ell_min` and `ell_max` fields appropriately.  If you wish to only
        skip `t`, `frame`, and `data`, you can simply use `WaveformBase.copy_without_data(W)`.  The latter is useful
        if, for example, you will only be making changes to those three fields, and want everything else preserved.

        Also note that some slicing operations can achieve similar -- but different -- goals.  For example,
        `w = w[:, :0]` will simply empty `data` and `ells`, without affecting the `time` and `frame`.

        """
        W = type(self)()
        state = copy.deepcopy(self.__dict__)
        state.pop('_WaveformBase__num')
        state.pop('t')
        state.pop('frame')
        state.pop('data')
        W.__dict__.update(state)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.copy_without_data()'.format(W, self))
        return W

    def _allclose(self, other, report_all=True, rtol=1e-10, atol=1e-10,
                  compare_history_beginnings=False, exceptions=[]):
        """Check that member data in two waveforms are the same

        For data sets (time, modes, etc.), the numpy function `np.allclose` is used, with the input tolerances.  See
        that function's documentation for more details.  The `*__num` datum is always ignored.  By default,
        the `history` is ignored, though this can be partially overridden -- in which case, the shortest subset of
        the histories is compared for exact equality.  This is probably only appropriate for the case where one
        waveform was created from the other.

        Parameters
        ----------
        other : object
            Another object subclassing WaveformBase to compare
        report_all: bool, optional
            Wait until all attributes have been checked (and reported on) before returning the verdict
        rtol : float, optional
            Relative tolerance to which to compare arrays (see np.allclose), defaults to 1e-10
        atol : float, optional
            Absolute tolerance to which to compare arrays (see np.allclose), defaults to 1e-10
        compare_history_beginnings: bool, optional
            Compare the shortest common part of the `history` fields for equality, defaults to False
        exceptions : list, optional
            Don't compare elements in this list, corresponding to keys in the object's `__dict__`, defaults to []

        """
        equality = True
        if not type(self) == type(other):  # not isinstance(other, self.__class__):
            warnings.warn("\n  (type(self)={0}) != (type(other)={1})".format(type(self), type(other)))
            equality = False
            if not report_all and not equality:
                return False
        for key, val in self.__dict__.items():
            if key.endswith('__num') or key in exceptions:
                continue
            elif key == 'history':
                if compare_history_beginnings:
                    min_length = min(len(self.history), len(other.history))
                    if self.history[:min_length] != other.history[:min_length]:
                        warnings.warn("\n  `history` fields differ")
                        equality = False
            elif isinstance(val, np.ndarray):
                if val.dtype == np.quaternion:
                    if not np.allclose(quaternion.as_float_array(val),
                                       quaternion.as_float_array(other.__dict__[key]), rtol, atol):
                        warnings.warn("\n  `{0}` fields differ".format(key))
                        equality = False
                elif not np.allclose(val, other.__dict__[key], rtol, atol):
                    warnings.warn("\n  `{0}` fields differ".format(key))
                    equality = False
            else:
                if not val == other.__dict__[key]:
                    warnings.warn("\n  (self.{0}={1}) != (other.{0}={2}) fields differ".format(key, val,
                                                                                               other.__dict__[key]))
                    equality = False
            if not report_all and not equality:
                return False
        return equality

    # Slicing
    @waveform_alterations
    def __getitem__(self, key):
        """Extract subsets of the data efficiently

        See the docstring of the WaveformBase class for examples.

        """
        W = WaveformBase.copy_without_data(self)

        # Remove trivial tuple structure first
        if isinstance(key, tuple) and len(key) == 1:
            key = key[0]

        # Now figure out which type of return is desired
        if isinstance(key, tuple) and 2 <= len(key) <= self.n_data_sets:
            # Return a subset of the data from a subset of times
            W.t = self.t[key[0]]
            W.frame = self.frame[key[0]]
            W.data = self.data[key]
        elif isinstance(key, slice) or isinstance(key, int):
            # Return complete data from a subset of times (key is slice), or
            # return complete data from a single instant in time (key is int)
            W.t = self.t[key]
            W.frame = self.frame[key]
            W.data = self.data[key]
        else:
            raise ValueError("Could not understand input `{0}` (of type `{1}`) ".format(key, type(key)))

        W.__history_depth__ -= 1
        W._append_history('{0} = {1}[{2}]'.format(W, self, key))

        return W


    @waveform_alterations
    def interpolate(self, tprime):
        """Interpolate the frame and data onto the new set of time steps

        Note that only `t`, `frame`, and `data` are changed in this function.  If there is a corresponding data set
        in a subclass, for example, the subclass must override this function to set that data set -- though this
        function should probably be called to handle the ugly stuff.

        """
        # Copy the information fields, but not the data
        W = WaveformBase.copy_without_data(self)

        W.t = np.copy(tprime)
        W.frame = quaternion.squad(self.frame, self.t, W.t)
        W.data = np.empty((W.n_times,)+self.data.shape[1:], dtype=self.data.dtype)
        if self.data.dtype == np.dtype(complex):
            for i in range(W.n_data_sets):
                W.data_2d[:, i] = (splev(W.t, splrep(self.t, self.data_2d[:, i].real, s=0), der=0)
                                   + 1j*splev(W.t, splrep(self.t, self.data_2d[:, i].imag, s=0), der=0))
        elif self.data.dtype == np.dtype(float):
            for i in range(W.n_data_sets):
                W.data_2d[:, i] = splev(W.t, splrep(self.t, self.data_2d[:, i], s=0), der=0)
        else:
            raise TypeError("Unknown self.data.dtype={0}".format(self.data.dtype))
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.interpolate({2})'.format(W, self, tprime))
        return W


    @waveform_alterations
    def SI_units(self, current_unit_mass_in_solar_masses, distance_from_source_in_megaparsecs=100):
        """Assuming current quantities are in geometric units, convert to SI units

        This function assumes that the `dataType`, `r_is_scaled_out`, and `m_is_scaled_out` attributes are correct,
        then scales the amplitude and time data appropriately so that the data correspond to data that could be
        observed from a source with the given total mass at the given distance.

        Note that the curvature scalars will have units of s^-2, rather than the arguably more correct m^-2.  This
        seems to be more standard in numerical relativity.  The result can be divided by `scipy.constants.c**2`
        to give units of m^-2 if desired.

        Parameters
        ----------
        current_unit_mass_in_solar_masses : float
            Mass of the system in the data converted to solar masses
        distance_from_source_in_megaparsecs : float, optional
            Output will be waveform as observed from this distance, default=100 (Mpc)

        """
        if not self.r_is_scaled_out:
            warning = ('\nTrying to convert to SI units, the radius is supposedly not scaled out.\n'
                       + 'This seems to suggest that the data may already be in some units...')
            warnings.warn(warning)
        if not self.m_is_scaled_out:
            warning = ('\nTrying to convert to SI units, the mass is supposedly not scaled out.\n'
                       + 'This seems to suggest that the data may already be in some units...')
            warnings.warn(warning)

        M_in_meters = current_unit_mass_in_solar_masses * m_sun_in_meters  # m
        M_in_seconds = M_in_meters / speed_of_light  # s
        R_in_meters = distance_from_source_in_megaparsecs * (1e6 * parsec_in_meters)  # m
        R_over_M = R_in_meters / M_in_meters  # [dimensionless]

        # The radius scaling `r_scaling` is the number of factors of the dimensionless quantity `R_over_M` required
        # to keep the waveform asymptotically constant.  So, for example, h and Psi4 both have `r_scaling=1`.  The
        # mass scaling `m_scaling` is the number of factors of `M_in_meters` required to make the waveform
        # dimensionless, and does not account for the factors of mass in the radius scale.  The Newman-Penrose
        # quantities are curvature quantities, so they have dimensions 1/m^2, and thus have `m_scaling=2`.
        if self.r_is_scaled_out:
            if self.m_is_scaled_out:
                amplitude_scaling = (R_over_M ** -self.r_scaling) * (M_in_meters ** -self.m_scaling)
            else:
                amplitude_scaling = R_over_M ** -self.r_scaling
        else:
            if self.m_is_scaled_out:
                amplitude_scaling = M_in_meters ** -self.m_scaling
            else:
                amplitude_scaling = 1.0

        # Copy the information fields, but not the data
        W = WaveformBase.copy_without_data(self)

        if self.m_is_scaled_out:
            W.t = M_in_seconds * self.t  # s
        else:
            W.t = np.copy(self.t)  # supposedly already in the appropriate units...
        W.frame = np.copy(self.frame)
        W.data = amplitude_scaling * self.data

        W.m_is_scaled_out = False
        W.r_is_scaled_out = False

        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.SI_units(current_unit_mass_in_solar_masses={2}, '
                          'distance_from_source_in_megaparsecs={3})'.format(W, self, current_unit_mass_in_solar_masses,
                                                                            distance_from_source_in_megaparsecs))

        return W





# Copyright (c) 2018, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import
import warnings
import numpy as np
import quaternion
import spherical_functions as sf
from scipy.interpolate import splev, splrep, InterpolatedUnivariateSpline, interp1d
import scipy.interpolate as interpolate

from . import Inertial, h
from .waveform_base import WaveformBase, waveform_alterations


class DiscreteSamplingMixin(object):
    @property
    def vartheta(self):
        return self._vartheta

    @vartheta.setter
    def vartheta(self, new_vartheta):
        raise AttributeError("Can't change vartheta.  This can only be done at initialization.")

    @property
    def varphi(self):
        return self._varphi

    @varphi.setter
    def varphi(self, new_varphi):
        raise AttributeError("Can't change varphi.  This can only be done at initialization.")

    @property
    def dt(self):
        return self._dt

    @dt.setter
    def dt(self, new_dt):
        raise AttributeError("Can't change dt.  This can only be done at initialization.")

    @property
    def sampling_frequency(self):
        return 1.0 / self.dt

    @sampling_frequency.setter
    def sampling_frequency(self, new_df):
        raise AttributeError("Can't change sampling_frequency.  This is a computed value.")

    @property
    def nyquist_frequency(self):
        return 0.5 / self.dt

    @nyquist_frequency.setter
    def nyquist_frequency(self, new_nf):
        raise AttributeError("Can't change nyquist_frequency.  This is a computed value.")


class WaveformInDetector(DiscreteSamplingMixin, WaveformBase):
    """Object containing the h+ and hx waveforms observed by a detector at a point

    Note that the data is stored as a complex quantity referred to as `h`, which represents `h+` - 1j*`hx`.  Note the
    minus sign.  The `hplus` and `hcross` components are given as properties of this object, which return the real
    part and negative imaginary part, respectively.  This unfortunate sign is due to the fact that `h` is generally
    defined as an object of spin-weight s=-2.

    Also note that the data may not actually represent `h`; it may represent `Psi4` or any other spin-weighted
    quantity.

    Finally, while the data stored are complex, gravitational-wave detectors do not (yet) measure both polarizations.
    This becomes important when taking the Fourier transform; usually the real part should be taken before applying
    the transform, to agree with standard data-analysis formulas.  On the other hand, both parts are required when
    trying to find the optimal match, because optimizing over the phase requires Fourier transforms of both parts.

    """

    def __init__(self, *args, **kwargs):
        """Initializer for WaveformInDetector object

        This is not a convenient function for constructing these objects.  See `WaveformInDetector.from_modes` for a
        more useful function.

        """
        self._vartheta = float(kwargs.pop('vartheta', 0.0))
        self._varphi = float(kwargs.pop('varphi', 0.0))
        self._dt = float(kwargs.pop('dt', 1./4096.))
        super(WaveformInDetector, self).__init__(*args, **kwargs)

    @classmethod
    def from_modes(cls, modes, vartheta=0.0, varphi=0.0, dt=1./4096., **kwargs):
        """Initializer for WaveformInDetector object given WaveformModes object

        Parameters
        ----------
        modes: WaveformModes object
            This is the object to be evaluated
        vartheta: float, defaults to 0.0
            Polar angle giving position of detector relative to the inertial frame of `modes`
        varphi: float, defaults to 0.0
            Azimuthal angle giving position of detector relative to the inertial frame of `modes`
        dt: float, defaults to 1/4096.
            This is the sampling time step (inverse sampling frequency) of the detector in seconds.  LIGO's sampling
            frequency is always a multiple of 1024 Hz.  Advanced LIGO acquires data at 16384 Hz, but downsamples for
            most purposes.  Also note that the Nyquist frequency (maximum unambiguous frequency) is half the sampling
            frequency.

        Optional keyword parameters
        ---------------------------
        sYlm_threshold: float, defaults to 1e-14
            If the value of the sYlm is below this threshold, the mode will not contribute to the result.  If the frame
            is different at each time step, the sYlm values must *all* be below this threshold.
        detector_response: complex, defaults to 1+0j
            The amplitude of this quantity multiplies the amplitude of the input, and the phase multiplies the phase.
            This quantity is usually denoted F = F+ + i * Fx, and is very closely related to the antenna pattern.

        """

        if modes.m_is_scaled_out or modes.r_is_scaled_out:
            raise ValueError("The input `modes` object has `m_is_scaled_out`={0}".format(modes.m_is_scaled_out)
                             + " and `r_is_scaled_out`={0}".format(modes.r_is_scaled_out)
                             + "\nDid you forget to transform to physical units with `modes.SI_units`?")

        n_steps = int((modes.t[-1]-modes.t[0])/dt)
        t = modes.t[0]+dt*np.arange(n_steps)
        data = np.zeros((n_steps,), dtype=np.complex)
        sYlm_threshold = float(kwargs.pop('sYlm_threshold', 1.0e-14))
        F = complex(kwargs.pop('detector_response', 1+0j))

        # Calling the interpolating functions too frequently will slow things down tremendously, because python is slow,
        # though they are fundamentally fortran routines.  So we can't call for each time step.  On the other hand,
        # interpolating all modes at each time step results in too much data -- typically 77 times as much as the result
        # of this function will be.  So it seems like the best option is to interpolate each mode separately, and add to
        # the result as we go.

        # Get a new look at that data, for easier interpolation
        mode_data = modes.data_2d.view(dtype=float)

        # Interpolate the frame to the given time steps
        if modes.frame.size == 0:
            frame = np.array([quaternion.one])
        elif modes.frame.size == 1:
            frame = np.copy(modes.frame)
        else:
            frame = quaternion.squad(modes.frame, modes.t, t)

        # Now get the rotor to be used in evaluating the SWSH [see Eq. (B4c) of arxiv:1409.4431]
        R = (~frame) * quaternion.from_spherical_coords(vartheta, varphi)

        # Loop over modes
        for i_m in range(modes.n_modes):
            # Calculate SWSH modes
            sYlm = F * sf.SWSH(R, modes.spin_weight, modes.LM[i_m])

            if np.max(np.abs(sYlm)) < sYlm_threshold:
                continue
            else:
                # Interpolate the modes to the needed time steps
                interpolated_mode = (InterpolatedUnivariateSpline(modes.t, mode_data[:, 2*i_m])(t)
                                     + 1j*InterpolatedUnivariateSpline(modes.t, mode_data[:, 2*i_m+1])(t))
                # Multiply by SWSH modes and add to data
                data += sYlm * interpolated_mode

        w = cls(vartheta=vartheta, varphi=varphi, dt=dt,
                t=t, frame=np.empty((0,), dtype=np.quaternion), data=data,
                history=modes.history+[''],
                frameType=Inertial, dataType=modes.dataType,
                r_is_scaled_out=False, m_is_scaled_out=False,
                constructor_statement=('WaveformInDetector.from_modes'
                    + '({0}, vartheta={1}, varphi={2}, dt={3}, '.format(modes, vartheta, varphi, dt)
                    + 'sYlm_threshold={0}, detector_response={1}'.format(sYlm_threshold, F)
                    + ', '.join(['{0}={1}'.format(key, val) for key, val in kwargs.items()]) + ')'))
        return w

    @property
    def h_plus(self):
        return self.data.real

    @property
    def h_cross(self):
        return -self.data.imag

    @waveform_alterations
    def ensure_validity(self, alter=True, assertions=False):
        """Try to ensure that the `WaveformInDetector` object is valid

        See `WaveformBase.ensure_validity` for the basic tests.  This function also includes a test that `t` and `dt`
        are consistent.

        """
        import numbers

        errors = []
        alterations = []

        if assertions:
            from .waveform_base import test_with_assertions
            test = test_with_assertions
        else:
            from .waveform_base import test_without_assertions
            test = test_without_assertions

        test(errors,
             np.allclose(self.t, self.t[0]+self.dt*np.arange(0, self.n_times), atol=1e-15, rtol=1e-15),
             'np.allclose(self.t, self.t[0]+self.dt*np.arange(0, self.n_times), atol=1e-15, rtol=1e-15) '
             + '\n# t[0]={0}, dt={1}, n_times={2}'.format(self.t[0], self.dt, self.n_times)
             + '\n# t={0}'.format(self.t)
        )
        test(errors,
             self.data.ndim == 1,
             'self.data.ndim == 1 # self.data.ndim={0}'.format(self.data.ndim))

        # Call the base class's version
        super(WaveformInDetector, self).ensure_validity(alter, assertions)

        self.__history_depth__ -= 1
        self._append_history('WaveformInDetector.ensure_validity' +
                             '({0}, alter={1}, assertions={2})'.format(self, alter, assertions))

        return True

    @waveform_alterations
    def copy_without_data(self):
        W = super(WaveformInDetector, self).copy_without_data()
        W._vartheta, W._varphi = self.vartheta, self.varphi
        W._dt = self.dt
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.copy_without_data()'.format(W, self))
        return W

    def __repr__(self):
        # "The goal of __str__ is to be readable; the goal of __repr__ is to be unambiguous." --- stackoverflow
        rep = super(WaveformInDetector, self).__repr__()
        rep += "\n# vartheta={0}, varphi={1}, dt={2}".format(self.vartheta, self.varphi, self.dt)
        return rep



from __future__ import print_function

import sys
from os import walk
from os.path import join, exists
from fnmatch import filter
from multiprocessing import Pool
from scri.SpEC import remove_avg_com_motion as racm


def run_in(args):
    (filename, i_this, i_tot) = args
    if exists(filename.replace('.h5', '_CoM.h5')):
        return
    print("Working on {0} -- {1} of {2}".format(filename, i_this, i_tot)); sys.stdout.flush()
    try:
        racm(filename + '/Extrapolated_N2.dir', plot=True, file_write_mode='w')
        try:
            racm(filename + '/Extrapolated_N3.dir', plot=True, file_write_mode='a')
        except:
            print('Failed in {0}/Extrapolated_N3.dir -- {1} of {2}'.format(filename, i_this, i_tot)); sys.stdout.flush()
        try:
            racm(filename + '/Extrapolated_N4.dir', plot=True, file_write_mode='a')
        except:
            print('Failed in {0}/Extrapolated_N4.dir -- {1} of {2}'.format(filename, i_this, i_tot)); sys.stdout.flush()
        try:
            racm(filename + '/OutermostExtraction.dir', plot=True, file_write_mode='a')
        except:
            print('Failed in {0}/OutermostExtraction.dir -- {1} of {2}'.format(filename, i_this, i_tot)); sys.stdout.flush()
    except:
        print('Failed in {0} -- {1} of {2}'.format(filename, i_this, i_tot)); sys.stdout.flush()
    print('Finished {0} -- {1} of {2}'.format(filename, i_this, i_tot)); sys.stdout.flush()


if __name__ == '__main__':
    print("Finding files to operate on")
    files = [join(root, 'rhOverM_Asymptotic_GeometricUnits.h5')
             for top_dir in ['Catalog', 'Incoming']
             for root, dirnames, filenames in walk(top_dir)
             for filename in filter(filenames, 'rhOverM_Asymptotic_GeometricUnits.h5')]
    print("Finished finding files to operate on")
    pool = Pool(processes=12)
    pool.map(run_in, zip(files, range(1, len(files)+1), [len(files)]*len(files)))



# Copyright (c) 2018, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

"""Submodule for operating on LVC-format waveform files"""

from .._version import __version__

from .file_io import (read_from_h5,)



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

"""Submodule for operating on SpEC waveform files"""

from .._version import __version__

from .com_motion import (com_motion, estimate_avg_com_motion, remove_avg_com_motion)
from .file_io import (read_from_h5, write_to_h5,)



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import

from math import sqrt

import numpy as np
import quaternion

from spherical_functions import ladder_operator_coefficient as ladder
from quaternion.numba_wrapper import jit, njit, xrange


@njit('void(c16[:,:], c16[:,:], i8[:,:], f8[:,:])')
def _LdtVector(data, datadot, lm, Ldt):
    """Helper function for the LdtVector function"""
    # Big, bad, ugly, obvious way to do the calculation
    # =================================================
    # L+ = Lx + i Ly      Lx =    (L+ + L-) / 2
    # L- = Lx - i Ly      Ly = -i (L+ - L-) / 2

    for i_mode in xrange(lm.shape[0]):
        L = lm[i_mode, 0]
        M = lm[i_mode, 1]
        for i_time in xrange(data.shape[0]):
            # Compute first in (+,-,z) basis
            Lp = (np.conjugate(data[i_time, i_mode + 1]) * datadot[i_time, i_mode] * ladder(L, M)
                  if M + 1 <= L
                  else 0.0 + 0.0j)
            Lm = (np.conjugate(data[i_time, i_mode - 1]) * datadot[i_time, i_mode] * ladder(L, -M)
                  if M - 1 >= -L
                  else 0.0 + 0.0j)
            Lz = np.conjugate(data[i_time, i_mode]) * datadot[i_time, i_mode] * M

            # Convert into (x,y,z) basis
            Ldt[i_time, 0] += 0.5 * (Lp.imag + Lm.imag)
            Ldt[i_time, 1] += -0.5 * (Lp.real - Lm.real)
            Ldt[i_time, 2] += Lz.imag
    return


def LdtVector(W):
    r"""Calculate the <Ldt> quantity with respect to the modes

    The vector is given in the (possibly rotating) mode frame (X,Y,Z),
    rather than the inertial frame (x,y,z).

    <Ldt>^{a} = \sum_{\ell,m,m'} \bar{f}^{\ell,m'} < \ell,m' | L_a | \ell,m > (df/dt)^{\ell,m}

    """
    Ldt = np.zeros((W.n_times, 3), dtype=float)
    _LdtVector(W.data, W.data_dot, W.LM, Ldt)
    return Ldt


@njit('void(c16[:,:], c16[:,:], i8[:,:], c16[:,:])')
def _LVector(data1, data2, lm, Lvec):
    """Helper function for the LVector function"""
    # Big, bad, ugly, obvious way to do the calculation
    # =================================================
    # L+ = Lx + i Ly      Lx =    (L+ + L-) / 2
    # L- = Lx - i Ly      Ly = -i (L+ - L-) / 2

    for i_mode in xrange(lm.shape[0]):
        L = lm[i_mode, 0]
        M = lm[i_mode, 1]
        for i_time in xrange(data1.shape[0]):
            # Compute first in (+,-,z) basis
            Lp = (np.conjugate(data1[i_time, i_mode + 1]) * data2[i_time, i_mode] * ladder(L, M)
                  if M + 1 <= L
                  else 0.0 + 0.0j)
            Lm = (np.conjugate(data1[i_time, i_mode - 1]) * data2[i_time, i_mode] * ladder(L, -M)
                  if M - 1 >= -L
                  else 0.0 + 0.0j)
            Lz = np.conjugate(data1[i_time, i_mode]) * data2[i_time, i_mode] * M

            # Convert into (x,y,z) basis
            Lvec[i_time, 0] += 0.5 * (Lp + Lm)
            Lvec[i_time, 1] += -0.5j * (Lp - Lm)
            Lvec[i_time, 2] += Lz
    return


def LVector(W1, W2):
    r"""Calculate the <L> quantity with respect to the modes

    The vector is given in the (possibly rotating) mode frame (X,Y,Z),
    rather than the inertial frame (x,y,z).

    <L>^{a} = \sum_{\ell,m,m'} \bar{f}^{\ell,m'} < \ell,m' | L_a | \ell,m > g^{\ell,m}

    """
    L = np.zeros((W1.n_times, 3), dtype=complex)
    _LVector(W1.data, W2.data, W1.LM, L)
    return L


@njit('void(c16[:,:], c16[:,:], i8[:,:], c16[:,:,:])')
def _LLComparisonMatrix(data1, data2, lm, LL):
    """Helper function for the LLComparisonMatrix function"""
    # Big, bad, ugly, obvious way to do the calculation
    # =================================================
    # L+ = Lx + i Ly      Lx =    (L+ + L-) / 2     Im(Lx) =  ( Im(L+) + Im(L-) ) / 2
    # L- = Lx - i Ly      Ly = -i (L+ - L-) / 2     Im(Ly) = -( Re(L+) - Re(L-) ) / 2
    # Lz = Lz             Lz = Lz                   Im(Lz) = Im(Lz)
    # LxLx =   (L+ + L-)(L+ + L-) / 4
    # LxLy = -i(L+ + L-)(L+ - L-) / 4
    # LxLz =   (L+ + L-)(  Lz   ) / 2
    # LyLx = -i(L+ - L-)(L+ + L-) / 4
    # LyLy =  -(L+ - L-)(L+ - L-) / 4
    # LyLz = -i(L+ - L-)(  Lz   ) / 2
    # LzLx =   (  Lz   )(L+ + L-) / 2
    # LzLy = -i(  Lz   )(L+ - L-) / 2
    # LzLz =   (  Lz   )(  Lz   )

    for i_mode in xrange(lm.shape[0]):
        L = lm[i_mode, 0]
        M = lm[i_mode, 1]
        for i_time in xrange(data1.shape[0]):
            # Compute first in (+,-,z) basis
            LpLp = (np.conjugate(data1[i_time, i_mode + 2]) * data2[i_time, i_mode] * (ladder(L, M + 1) * ladder(L, M))
                    if M + 2 <= L
                    else 0.0 + 0.0j)
            LpLm = (np.conjugate(data1[i_time, i_mode]) * data2[i_time, i_mode] * (ladder(L, M - 1) * ladder(L, -M))
                    if M - 1 >= -L
                    else 0.0 + 0.0j)
            LmLp = (np.conjugate(data1[i_time, i_mode]) * data2[i_time, i_mode] * (ladder(L, -(M + 1)) * ladder(L, M))
                    if M + 1 <= L
                    else 0.0 + 0.0j)
            LmLm = (
                np.conjugate(data1[i_time, i_mode - 2]) * data2[i_time, i_mode] * (ladder(L, -(M - 1)) * ladder(L, -M))
                if M - 2 >= -L
                else 0.0 + 0.0j)
            LpLz = (np.conjugate(data1[i_time, i_mode + 1]) * data2[i_time, i_mode] * (ladder(L, M) * M)
                    if M + 1 <= L
                    else 0.0 + 0.0j)
            LzLp = (np.conjugate(data1[i_time, i_mode + 1]) * data2[i_time, i_mode] * ((M + 1) * ladder(L, M))
                    if M + 1 <= L
                    else 0.0 + 0.0j)
            LmLz = (np.conjugate(data1[i_time, i_mode - 1]) * data2[i_time, i_mode] * (ladder(L, -M) * M)
                    if M - 1 >= -L
                    else 0.0 + 0.0j)
            LzLm = (np.conjugate(data1[i_time, i_mode - 1]) * data2[i_time, i_mode] * ((M - 1) * ladder(L, -M))
                    if M - 1 >= -L
                    else 0.0 + 0.0j)
            LzLz = np.conjugate(data1[i_time, i_mode]) * data2[i_time, i_mode] * M ** 2

            # Convert into (x,y,z) basis
            LL[i_time, 0, 0] += 0.25 * (LpLp + LmLm + LmLp + LpLm)
            LL[i_time, 0, 1] += -0.25j * (LpLp - LmLm + LmLp - LpLm)
            LL[i_time, 0, 2] += 0.5 * (LpLz + LmLz)
            LL[i_time, 1, 0] += -0.25j * (LpLp - LmLp + LpLm - LmLm)
            LL[i_time, 1, 1] += -0.25 * (LpLp - LmLp - LpLm + LmLm)
            LL[i_time, 1, 1] += -0.5j * (LpLz - LmLz)
            LL[i_time, 2, 0] += 0.5 * (LzLp + LzLm)
            LL[i_time, 2, 1] += -0.5j * (LzLp - LzLm)
            LL[i_time, 2, 2] += LzLz

            # # Symmetrize
            # LL[i_time,0,0] += ( LxLx ).real
            # LL[i_time,0,1] += ( LxLy + LyLx ).real/2.0
            # LL[i_time,0,2] += ( LxLz + LzLx ).real/2.0
            # LL[i_time,1,0] += ( LyLx + LxLy ).real/2.0
            # LL[i_time,1,1] += ( LyLy ).real
            # LL[i_time,1,2] += ( LyLz + LzLy ).real/2.0
            # LL[i_time,2,0] += ( LzLx + LxLz ).real/2.0
            # LL[i_time,2,1] += ( LzLy + LyLz ).real/2.0
            # LL[i_time,2,2] += ( LzLz ).real
    return


def LLComparisonMatrix(W1, W2):
    r"""Calculate the <LL> quantity with respect to the modes of two Waveforms

    The matrix is given in the (possibly rotating) mode frame (X,Y,Z),
    rather than the inertial frame (x,y,z).

    <LL>^{ab} = \sum_{\ell,m,m'} \bar{f}^{\ell,m'} < \ell,m' | L_a L_b | \ell,m > g^{\ell,m}

    """
    LL = np.zeros((W1.n_times, 3, 3), dtype=complex)
    _LLComparisonMatrix(W1.data, W2.data, W1.LM, LL)
    return LL


@njit('void(c16[:,:], i8[:,:], f8[:,:,:])')
def _LLMatrix(data, lm, LL):
    """Helper function for the LLMatrix function"""
    # Big, bad, ugly, obvious way to do the calculation
    # =================================================
    # L+ = Lx + i Ly      Lx =    (L+ + L-) / 2     Im(Lx) =  ( Im(L+) + Im(L-) ) / 2
    # L- = Lx - i Ly      Ly = -i (L+ - L-) / 2     Im(Ly) = -( Re(L+) - Re(L-) ) / 2
    # Lz = Lz             Lz = Lz                   Im(Lz) = Im(Lz)
    # LxLx =   (L+ + L-)(L+ + L-) / 4
    # LxLy = -i(L+ + L-)(L+ - L-) / 4
    # LxLz =   (L+ + L-)(  Lz   ) / 2
    # LyLx = -i(L+ - L-)(L+ + L-) / 4
    # LyLy =  -(L+ - L-)(L+ - L-) / 4
    # LyLz = -i(L+ - L-)(  Lz   ) / 2
    # LzLx =   (  Lz   )(L+ + L-) / 2
    # LzLy = -i(  Lz   )(L+ - L-) / 2
    # LzLz =   (  Lz   )(  Lz   )

    for i_mode in xrange(lm.shape[0]):
        L = lm[i_mode, 0]
        M = lm[i_mode, 1]
        for i_time in xrange(data.shape[0]):
            # Compute first in (+,-,z) basis
            LpLp = (np.conjugate(data[i_time, i_mode + 2]) * data[i_time, i_mode] * (ladder(L, M + 1) * ladder(L, M))
                    if M + 2 <= L
                    else 0.0 + 0.0j)
            LpLm = (np.conjugate(data[i_time, i_mode]) * data[i_time, i_mode] * (ladder(L, M - 1) * ladder(L, -M))
                    if M - 1 >= -L
                    else 0.0 + 0.0j)
            LmLp = (np.conjugate(data[i_time, i_mode]) * data[i_time, i_mode] * (ladder(L, -(M + 1)) * ladder(L, M))
                    if M + 1 <= L
                    else 0.0 + 0.0j)
            LmLm = (
                np.conjugate(data[i_time, i_mode - 2]) * data[i_time, i_mode] * (ladder(L, -(M - 1)) * ladder(L, -M))
                if M - 2 >= -L
                else 0.0 + 0.0j)
            LpLz = (np.conjugate(data[i_time, i_mode + 1]) * data[i_time, i_mode] * (ladder(L, M) * M)
                    if M + 1 <= L
                    else 0.0 + 0.0j)
            LzLp = (np.conjugate(data[i_time, i_mode + 1]) * data[i_time, i_mode] * ((M + 1) * ladder(L, M))
                    if M + 1 <= L
                    else 0.0 + 0.0j)
            LmLz = (np.conjugate(data[i_time, i_mode - 1]) * data[i_time, i_mode] * (ladder(L, -M) * M)
                    if M - 1 >= -L
                    else 0.0 + 0.0j)
            LzLm = (np.conjugate(data[i_time, i_mode - 1]) * data[i_time, i_mode] * ((M - 1) * ladder(L, -M))
                    if M - 1 >= -L
                    else 0.0 + 0.0j)
            LzLz = np.conjugate(data[i_time, i_mode]) * data[i_time, i_mode] * M ** 2

            # Convert into (x,y,z) basis
            LxLx = 0.25 * (LpLp + LmLm + LmLp + LpLm)
            LxLy = -0.25j * (LpLp - LmLm + LmLp - LpLm)
            LxLz = 0.5 * (LpLz + LmLz)
            LyLx = -0.25j * (LpLp - LmLp + LpLm - LmLm)
            LyLy = -0.25 * (LpLp - LmLp - LpLm + LmLm)
            LyLz = -0.5j * (LpLz - LmLz)
            LzLx = 0.5 * (LzLp + LzLm)
            LzLy = -0.5j * (LzLp - LzLm)
            # LzLz = (LzLz)

            # Symmetrize
            LL[i_time, 0, 0] += LxLx.real
            LL[i_time, 0, 1] += (LxLy + LyLx).real / 2.0
            LL[i_time, 0, 2] += (LxLz + LzLx).real / 2.0
            LL[i_time, 1, 0] += (LyLx + LxLy).real / 2.0
            LL[i_time, 1, 1] += LyLy.real
            LL[i_time, 1, 2] += (LyLz + LzLy).real / 2.0
            LL[i_time, 2, 0] += (LzLx + LxLz).real / 2.0
            LL[i_time, 2, 1] += (LzLy + LyLz).real / 2.0
            LL[i_time, 2, 2] += LzLz.real
    return


def LLMatrix(W):
    r"""Calculate the <LL> quantity with respect to the modes

    The matrix is given in the (possibly rotating) mode frame (X,Y,Z),
    rather than the inertial frame (x,y,z).

    This quantity is as prescribed by O'Shaughnessy et al.
    <http://arxiv.org/abs/1109.5224>, except that no normalization is
    imposed, and this operates on whatever type of data is input.

    <LL>^{ab} = Re \{ \sum_{\ell,m,m'} \bar{f}^{\ell,m'} < \ell,m' | L_a L_b | \ell,m > f^{\ell,m} \}

    """
    LL = np.zeros((W.n_times, 3, 3), dtype=float)
    _LLMatrix(W.data, W.LM, LL)
    return LL


@njit('void(f8[:,:], f8[:], i8)')
def _LLDominantEigenvector(dpa, dpa_i, i_index):
    """Jitted helper function for LLDominantEigenvector"""
    # Make the initial direction closer to RoughInitialEllDirection than not
    if (dpa_i[0] * dpa[i_index, 0] + dpa_i[1] * dpa[i_index, 1] + dpa_i[2] * dpa[i_index, 2]) < 0.:
        dpa[i_index, 0] *= -1
        dpa[i_index, 1] *= -1
        dpa[i_index, 2] *= -1
    # Now, go through and make the vectors reasonably continuous.
    d = -1
    LastNorm = sqrt(dpa[i_index, 0] ** 2 + dpa[i_index, 1] ** 2 + dpa[i_index, 2] ** 2)
    for i in xrange(i_index-1, -1, -1):
        Norm = dpa[i, 0] ** 2 + dpa[i, 1] ** 2 + dpa[i, 2] ** 2
        dNorm = ((dpa[i, 0] - dpa[i - d, 0]) ** 2 + (dpa[i, 1] - dpa[i - d, 1]) ** 2
                 + (dpa[i, 2] - dpa[i - d, 2]) ** 2)
        if dNorm > Norm:
            dpa[i, 0] *= -1
            dpa[i, 1] *= -1
            dpa[i, 2] *= -1
        # While we're here, let's just normalize that last one
        if LastNorm != 0.0 and LastNorm != 1.0:
            dpa[i - d, 0] /= LastNorm
            dpa[i - d, 1] /= LastNorm
            dpa[i - d, 2] /= LastNorm
        LastNorm = sqrt(Norm)
    if LastNorm != 0.0 and LastNorm != 1.0:
        dpa[0, 0] /= LastNorm
        dpa[0, 1] /= LastNorm
        dpa[0, 2] /= LastNorm
    d = 1
    LastNorm = sqrt(dpa[i_index, 0] ** 2 + dpa[i_index, 1] ** 2 + dpa[i_index, 2] ** 2)
    for i in xrange(i_index+1, dpa.shape[0]):
        Norm = dpa[i, 0] ** 2 + dpa[i, 1] ** 2 + dpa[i, 2] ** 2
        dNorm = ((dpa[i, 0] - dpa[i - d, 0]) ** 2 + (dpa[i, 1] - dpa[i - d, 1]) ** 2
                 + (dpa[i, 2] - dpa[i - d, 2]) ** 2)
        if dNorm > Norm:
            dpa[i, 0] *= -1
            dpa[i, 1] *= -1
            dpa[i, 2] *= -1
        # While we're here, let's just normalize that last one
        if LastNorm != 0.0 and LastNorm != 1.0:
            dpa[i - d, 0] /= LastNorm
            dpa[i - d, 1] /= LastNorm
            dpa[i - d, 2] /= LastNorm
        LastNorm = sqrt(Norm)
    if LastNorm != 0.0 and LastNorm != 1.0:
        dpa[-1, 0] /= LastNorm
        dpa[-1, 1] /= LastNorm
        dpa[-1, 2] /= LastNorm
    return


def LLDominantEigenvector(W, RoughDirection=np.array([0.0, 0.0, 1.0]), RoughDirectionIndex=0):
    """Calculate the principal axis of the LL matrix

    \param Lmodes L modes to evaluate (optional)
    \param RoughDirection Vague guess about the preferred initial (optional)

    If Lmodes is empty (default), all L modes are used.  Setting
    Lmodes to [2] or [2,3,4], for example, restricts the range of
    the sum.

    Ell is the direction of the angular velocity for a PN system, so
    some rough guess about that direction allows us to choose the
    direction of the eigenvectors output by this function to be more
    parallel than anti-parallel to that direction.  The default is
    to simply choose the z axis, since this is most often the
    correct choice anyway.

    The vector is given in the (possibly rotating) mode frame
    (X,Y,Z), rather than the inertial frame (x,y,z).

    """
    # Calculate the LL matrix at each instant
    LL = np.zeros((W.n_times, 3, 3), dtype=float)
    _LLMatrix(W.data, W.LM, LL)

    # This is the eigensystem
    eigenvals, eigenvecs = np.linalg.eigh(LL)

    # Now we find and normalize the dominant principal axis at each
    # moment, made continuous
    dpa = eigenvecs[:, :, 2]  # `eigh` always returns eigenvals in *increasing* order
    _LLDominantEigenvector(dpa, RoughDirection, RoughDirectionIndex)

    return dpa


#@jit
def angular_velocity(W, include_frame_velocity=False):
    """Angular velocity of Waveform

    This function calculates the angular velocity of a Waveform object from
    its modes.  This was introduced in Sec. II of "Angular velocity of
    gravitational radiation and the corotating frame"
    <http://arxiv.org/abs/1302.2919>.  Essentially, this is the angular
    velocity of the rotating frame in which the time dependence of the modes
    is minimized.

    The vector is given in the (possibly rotating) mode frame (X,Y,Z),
    which is not necessarily equal to the inertial frame (x,y,z).

    """

    # Calculate the <Ldt> vector and <LL> matrix at each instant
    l = W.LdtVector()
    ll = W.LLMatrix()

    # Solve <Ldt> = - <LL> . omega
    omega = -np.linalg.solve(ll, l)

    if include_frame_velocity and len(W.frame)==W.n_times:
        from quaternion import derivative, as_quat_array, as_float_array
        Rdot = as_quat_array(derivative(as_float_array(W.frame), W.t))
        omega += as_float_array(2 * Rdot * W.frame.conjugate())[:, 1:]

    return omega


#def corotating_frame(W, Ri=quaternion.one, ti=0.0, tolerance=1e-12):
def corotating_frame(W, R0=quaternion.one, tolerance=1e-12, z_alignment_region=None, return_omega=False):
    """Return rotor taking current mode frame into corotating frame

    This function simply evaluates the angular velocity of the waveform, and
    then integrates it to find the corotating frame itself.  This frame is
    defined to be the frame in which the time-dependence of the waveform is
    minimized --- at least, to the extent possible with a time-dependent
    rotation.  This frame is only unique up to a single overall rotation, which
    is passed in as an optional argument to this function.

    Parameters
    ----------
    W: Waveform
    R0: quaternion [defaults to 1]
        Value of the output rotation at the first output instant
    tolerance: float [defaults to 1e-12]
        Absolute tolerance used in integration
    z_alignment_region: None or 2-tuple of floats [defaults to None]
        If not None, the dominant eigenvector of the <LL> matrix is aligned with the z axis,
        averaging over this portion of the data.  The first and second elements of the input are
        considered fractions of the inspiral at which to begin and end the average.  For example,
        (0.1, 0.9) would lead to starting 10% of the time from the first time step to the max norm
        time, and ending at 90% of that time.
    return_omega: bool [defaults to False]
        If True, return a 2-tuple consisting of the frame (the usual returned object) and the
        angular-velocity data.  That is frequently also needed, so this is just a more efficient
        way of getting the data.

    """
    from quaternion.quaternion_time_series import integrate_angular_velocity, squad
    omega = angular_velocity(W)
    t, frame = integrate_angular_velocity((W.t, omega), t0=W.t[0], t1=W.t[-1], R0=R0, tolerance=tolerance)
    if z_alignment_region is None:
        correction_rotor = quaternion.one
    else:
        initial_time = W.t[0]
        inspiral_time = W.max_norm_time() - initial_time
        t1 = initial_time + z_alignment_region[0]*inspiral_time
        t2 = initial_time + z_alignment_region[1]*inspiral_time
        i1 = np.argmin(np.abs(W.t-t1))
        i2 = np.argmin(np.abs(W.t-t2))
        R = frame[i1:i2]
        i1m = max(0, i1-10)
        i1p = i1m + 21
        RoughDirection = W[i1m:i1p].angular_velocity()[10]
        Vhat = W[i1:i2].LLDominantEigenvector(RoughDirection=RoughDirection, RoughDirectionIndex=0)
        Vhat_corot = np.array([(Ri.conjugate() * quaternion.quaternion(*Vhati) * Ri).vec
                               for Ri, Vhati in zip(R, Vhat)])
        Vhat_corot_mean = quaternion.quaternion(*np.mean(Vhat_corot, axis=0)).normalized()
        # print(i1, i2, i1m, i1p, RoughDirection, Vhat[0], Vhat_corot[0], Vhat_corot_mean)
        correction_rotor = np.sqrt_of_rotor(-quaternion.z * Vhat_corot_mean).inverse()
    # R = squad(R, t, W.t)
    if return_omega:
        return (frame * correction_rotor, omega)
    else:
        return frame * correction_rotor



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import

from .waveform_base import WaveformBase, waveform_alterations

import warnings
import numpy as np
import spherical_functions as sf
from . import *


class WaveformModes(WaveformBase):
    """Object containing time, frame, and data, along with related information

    This object collects all the data needed to manipulate time-series of spin-weighted spherical-harmonic modes of
    gravitational waves, as well as a few useful informational members.  Various methods exist to manipulate the data
    and extract important information.

    As much as possible, completeness of modes is enforced -- meaning that if any (ell,m) mode with a given ell is
    included in the data, all modes with abs(m)<=ell must also be included.  The data are stored as complex numbers,
    and assumed to be in standard order corresponding to

        [f(ell,m) for ell in range(ell_min, ell_max+1) for m in range(-ell,ell+1)]

    There is no automated check that can be done to assure that the order is correct.  (Precession breaks all
    symmetries, for example.)  The only check that is done ensures that the input data array has the correct
    dimensions.  It is up to the user to ensure that the order is correct.

    Attributes
    ----------
    t : float array
        Time steps corresponding to other data
    frame : quaternion array
        Rotors taking static basis onto decomposition basis
    data : 2-d complex array
        Mode values of the spin-weighted spherical-harmonic decomposition. The array is assumed to have a first
        dimension equal to the length of `t`, and a second dimension equal to the number of modes as deduced from the
        values of ell_min and ell_max below.  As noted above, a very particular order is assumed for the data.
    ell_min : int
        Smallest ell value present in `data`.
    ell_max : int
        Largest ell value present in `data`.
    LM : int array (read only)
        Array of (ell,m) values corresponding to the `data` member.  This is automatically constructed based on the
        values of ell_min and ell_max, and cannot be reassigned.
    history : list of strings
        As far as possible, all functions applied to the object are recorded in the `history` variable.  In fact,
        the object should almost be able to be recreated using the commands in the history list. Commands taking
        large arrays, however, are shortened -- so the data will not be entirely reconstructable.
    frameType : int
        Index corresponding to `scri.FrameType` appropriate for `data`.
    dataType : int
        Index corresponding to `scri.DataType` appropriate for `data`.
    r_is_scaled_out : bool
        True if the `data` have been multiplied by the appropriate power of radius so that the asymptotic value can
        be finite and nonzero.
    m_is_scaled_out : bool
        True if the `data` have been scaled by the appropriate value of the total mass so that they are dimensionless.
    num : int (read only)
        Automatically assigned number of this object.  The constructor of this type keeps count of the number of
        objects it has created, to assign each object a more-or-less unique ID for use in the history strings.  This
        counter is reset at the beginning of each python session.  Subclasses should automatically have a different
        counter.

    Indexing
    --------
    WaveformMode objects can be indexed much like a numpy array, where the first dimension gives the time indices,
    and the second gives the mode indices. This will return another WaveformMode object containing slices of the
    original data.

    It is important to note, however, that as with numpy array slices, slicing a WaveformMode object will not
    typically copy the original data; the result will simply be a view into the data.  This means that changing the
    data in the slice can change the data in the original.  If you want to make a copy, you should probably use the
    copy constructor: `W2 = WaveformMode(W1)`. It is also possible to use the standard copy.deepcopy method.

    Also note that the first slice dimension corresponds to the indices of the time data, but the second dimension
    does NOT correspond to indices.  Instead, because this object tries to enforce completeness of the mode data,
    you can only slice a WaveformBase with respect to modes.  If you really want just some of the data, you'll need
    to extract it as a plain array, and then just slice as usual.

    For example,

    >>> W  = WaveformModes()
    >>> W[10:-20]

    will give all modes in the data, but only at times starting with the 10th time step, and ending one before the
    -20th time step.  Meanwhile,

    >>> W[10:-20,2]

    will give the same range of times, but only the ell=2 modes -- which includes (ell,m) modes (2,-2) through (2, 2).
    Similarly,

    >>> W[10:-20,2:5]

    will return the same range of times, along with the ell=2,3,4 modes. Note the lack of ell=5 mode, for consistency
    with python's usual slice syntax.

    >>> W[:,:0]

    will return all time steps, along with all `frame` data, but the `lm` data and `data` will be empty (because the
    `:0` term selects everything before the 0th element).  Similarly,

    >>> W[:0,:0]

    is empty of all the numerical data.



    """

    def __init__(self, *args, **kwargs):
        """Initializer for WaveformModes object

        This initializer is primarily a wrapper around the WaveformBase initializer.  See the docstring of
        WaveformBase for more details.  The only difference in calling is that this takes two additional keyword
        parameters:

        Keyword parameters
        ------------------
        ell_min : int, defaults to 0
        ell_max : int, defaults to -1

        """
        if len(args) == 1 and isinstance(args[0], type(self)):
            other = args[0]
            # Do not directly access __ell_min, __ell_max, or __LM outside of this initializer function; use ell_min,
            #  ell_max, or LM instead
            self.__ell_min = other.__ell_min
            self.__ell_max = other.__ell_max
            self.__LM = np.copy(other.__LM)
        else:
            # Do not directly access __ell_min, __ell_max, or __LM outside of this initializer function; use ell_min,
            # ell_max, or LM instead
            self.__ell_min = kwargs.pop('ell_min', 0)
            self.__ell_max = kwargs.pop('ell_max', -1)
            self.__LM = sf.LM_range(self.__ell_min, self.__ell_max)
        super(WaveformModes, self).__init__(*args, **kwargs)

    @waveform_alterations
    def ensure_validity(self, alter=True, assertions=False):
        """Try to ensure that the `WaveformModes` object is valid

        See `WaveformBase.ensure_validity` for the basic tests.  This function also includes tests that `data` is
        complex, and consistent with the ell_min and ell_max values.

        """
        import numbers

        errors = []
        alterations = []

        if assertions:
            from .waveform_base import test_with_assertions
            test = test_with_assertions
        else:
            from .waveform_base import test_without_assertions
            test = test_without_assertions

        # We first need to check that the ell values make sense,
        # because we'll use these below
        test(errors,
             isinstance(self.__ell_min, numbers.Integral),
             'isinstance(self.__ell_min, numbers.Integral) # type(self.__ell_min)={0}'.format(type(self.__ell_min)))
        test(errors,
             isinstance(self.__ell_max, numbers.Integral),
             'isinstance(self.__ell_max, numbers.Integral) # type(self.__ell_max)={0}'.format(type(self.__ell_max)))
        test(errors,
             self.__ell_min >= 0,
             'self.__ell_min>=0 # {0}'.format(self.__ell_min))
        test(errors,
             self.__ell_max >= self.__ell_min - 1,
             'self.__ell_max>=self.__ell_min-1 # self.__ell_max={0}; self.__ell_min-1={1}'.format(self.__ell_max,
                                                                                                  self.__ell_min - 1))
        if alter and not np.array_equal(self.__LM, sf.LM_range(self.ell_min, self.ell_max)):
            self.__LM = sf.LM_range(self.ell_min, self.ell_max)
            alterations += [
                '{0}._{1}__LM = sf.LM_range({2}, {3})'.format(self, type(self).__name__, self.ell_min, self.ell_max)]
        test(errors,
             np.array_equal(self.__LM, sf.LM_range(self.ell_min, self.ell_max)),
             'np.array_equal(self.__LM, sf.LM_range(self.ell_min, self.ell_max))')

        test(errors,
             self.data.dtype == np.dtype(np.complex),
             'self.data.dtype == np.dtype(np.complex) # self.data.dtype={0}'.format(self.data.dtype))
        test(errors,
             self.data.ndim >= 2,
             'self.data.ndim >= 2 # self.data.ndim={0}'.format(self.data.ndim))
        test(errors,
             self.data.shape[1] == self.__LM.shape[0],
             'self.data.shape[1]==self.__LM.shape[0] '
             '# self.data.shape={0}; self.__LM.shape[0]={1}'.format(self.data.shape[1], self.__LM.shape[0]))

        if alterations:
            self._append_history(alterations)
            print("The following alterations were made:\n\t" + '\n\t'.join(alterations))
        if errors:
            print("The following conditions were found to be incorrectly False:\n\t" + '\n\t'.join(errors))
            return False

        # Call the base class's version
        super(WaveformModes, self).ensure_validity(alter, assertions)

        self.__history_depth__ -= 1
        self._append_history('WaveformModes.ensure_validity' +
                             '({0}, alter={1}, assertions={2})'.format(self, alter, assertions))

        return True

    # Mode information
    @property
    def n_modes(self):
        return self.data.shape[1]

    @property
    def ell_min(self):
        return self.__ell_min

    @ell_min.setter
    def ell_min(self, new_ell_min):
        self.__ell_min = new_ell_min
        self.__LM = sf.LM_range(self.ell_min, self.ell_max)
        if self.n_modes != self.__LM.shape[0]:
            warning = "\nWaveform's data.shape={0} does not agree with ".format(self.data.shape) \
                      + "(ell_min,ell_max)=({0},{1}).\n".format(self.ell_min, self.ell_max) \
                      + "Hopefully you are about to reset `data`.  To suppress this warning,\n" \
                      + "reset `data` before resetting ell_min and/or ell_max."
            warnings.warn(warning)

    @property
    def ell_max(self):
        return self.__ell_max

    @ell_max.setter
    def ell_max(self, new_ell_max):
        self.__ell_max = new_ell_max
        self.__LM = sf.LM_range(self.ell_min, self.ell_max)
        if self.n_modes != self.__LM.shape[0]:
            warning = "\nWaveform's data.shape={0} does not agree with ".format(self.data.shape) \
                      + "(ell_min,ell_max)=({0},{1}).\n".format(self.ell_min, self.ell_max) \
                      + "Hopefully you are about to reset `data`.  To suppress this warning,\n" \
                      + "reset `data` before resetting ell_min and/or ell_max."
            warnings.warn(warning)

    @property
    def ells(self):
        """Return self.ell_min,self.ell_max"""
        return self.__ell_min, self.__ell_max

    @ells.setter
    def ells(self, new_ells):
        """Setting both at once can be necessary when changing the shape of `data`"""
        self.__ell_min = new_ells[0]
        self.__ell_max = new_ells[1]
        self.__LM = sf.LM_range(self.ell_min, self.ell_max)
        if self.n_modes != self.__LM.shape[0]:
            warning = "\nWaveform's data.shape={0} does not agree with ".format(self.data.shape) \
                      + "(ell_min,ell_max)=({0},{1}).\n".format(self.ell_min, self.ell_max) \
                      + "Hopefully you are about to reset `data`.  To avoid this warning,\n" \
                      + "reset `data` before resetting ell_min and/or ell_max."
            warnings.warn(warning)

    @property
    def LM(self):
        """Array of (ell,m) values in Waveform data

        This array is just a flat array of `[ell,m]` pairs.  It is
        automatically recomputed each time `ell_min` or `ell_max` is
        changed.  Specifically, it is

            np.array([[ell,m]
                      for ell in range(self.ell_min,self.ell_max+1)
                      for m in range(-ell,ell+1)])

        """
        return self.__LM

    @LM.setter
    def LM(self, newLM):
        raise AttributeError("Can't set LM data.  This can only be controlled through the ell_min and ell_max members.")

    def index(self, ell, m):
        """Index of given (ell,m) mode in the data

        Parameters
        ----------
        ell: int
        m: int

        Returns
        -------
        idx: int
            Index such that self.LM[idx] is [ell, m]

        """
        return sf.LM_index(ell, m, self.ell_min)

    def indices(self, ell_m):
        """Indices of given (ell,m) modes in the data

        Parameters
        ----------
        ell_m: Nx2 array-like, int

        Returns
        -------
        idcs: array of int
            Index such that self.LM[idcs] is `ell_m`

        """
        ell_m = np.array(ell_m, copy=False)
        if not (ell_m.dtype == np.int and ell_m.shape[1] == 2):
            raise ValueError("Input `ell_m` should be an Nx2 sequence of integers")
        return ell_m[:, 0] * (ell_m[:, 0] + 1) - self.ell_min ** 2 + ell_m[:, 1]

    # Involutions
    @property
    @waveform_alterations
    def x_parity_conjugate(self):
        """Reflect modes across y-z plane (along x axis)

        See "Gravitational-wave modes from precessing black-hole binaries" by
        Boyle et al. (2014) for more details.

        """
        if self.dataType == UnknownDataType:
            raise ValueError("Cannot compute parity type for {0}.".format(self.data_type_string))
        W = self[:, :0]  # W without `data`, and `ells`=(0,-1); different from `self.copy_without_data()`
        W.data = np.empty_like(self.data)
        W.ells = self.ells
        W.frame = np.x_parity_conjugate(self.frame)
        for ell in range(W.ell_min, W.ell_max + 1):
            lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1) if (m % 2) == 0]
            W.data[:, lm_indices] = np.conjugate(self.data[:, lm_indices])
            lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1) if (m % 2) != 0]
            W.data[:, lm_indices] = -np.conjugate(self.data[:, lm_indices])
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.x_parity_conjugate'.format(W, self))
        return W

    @property
    @waveform_alterations
    def x_parity_symmetric_part(self):
        """Return component of waveform invariant under `x_parity_conjugate`"""
        W = self.x_parity_conjugate
        W.data = 0.5 * (self.data + W.data)
        W.frame = np.x_parity_symmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.x_parity_symmetric_part'.format(W, self))
        return W

    @property
    @waveform_alterations
    def x_parity_antisymmetric_part(self):
        """Return component of waveform that changes sign under `x_parity_conjugate`"""
        W = self.x_parity_conjugate
        W.data = 0.5 * (self.data - W.data)
        W.frame = np.x_parity_antisymmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.x_parity_antisymmetric_part'.format(W, self))
        return W

    @property
    def x_parity_violation_squared(self):
        """(Squared) norm of x-parity-antisymmetric component of waveform"""
        return self.x_parity_antisymmetric_part.norm()

    @property
    def x_parity_violation_normalized(self):
        """Norm of x-parity-antisymmetric component divided by norm of waveform"""
        return np.sqrt(self.x_parity_antisymmetric_part.norm() / self.norm())

    @property
    @waveform_alterations
    def y_parity_conjugate(self):
        """Reflect modes across x-z plane (along y axis)

        See "Gravitational-wave modes from precessing black-hole binaries" by
        Boyle et al. (2014) for more details.

        """
        if self.dataType == UnknownDataType:
            raise ValueError("Cannot compute parity type for {0}.".format(self.data_type_string))
        W = self[:, :0]  # W without `data`, and `ells`=(0,-1); different from `self.copy_without_data()`
        W.data = np.conjugate(self.data)
        W.ells = self.ells
        W.frame = np.y_parity_conjugate(self.frame)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.y_parity_conjugate'.format(W, self))
        return W

    @property
    @waveform_alterations
    def y_parity_symmetric_part(self):
        """Component of waveform invariant under `y_parity_conjugate`"""
        W = self.y_parity_conjugate
        W.data = 0.5 * (self.data + W.data)
        W.frame = np.y_parity_symmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.y_parity_symmetric_part'.format(W, self))
        return W

    @property
    @waveform_alterations
    def y_parity_antisymmetric_part(self):
        """Component of waveform that changes sign under `y_parity_conjugate`"""
        W = self.y_parity_conjugate
        W.data = 0.5 * (self.data - W.data)
        W.frame = np.y_parity_antisymmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.y_parity_antisymmetric_part'.format(W, self))
        return W

    @property
    def y_parity_violation_squared(self):
        """(Squared) norm of y-parity-antisymmetric component of waveform"""
        return self.y_parity_antisymmetric_part.norm()

    @property
    def y_parity_violation_normalized(self):
        """Norm of y-parity-antisymmetric component divided by norm of waveform"""
        return np.sqrt(self.y_parity_antisymmetric_part.norm() / self.norm())

    @property
    @waveform_alterations
    def z_parity_conjugate(self):
        """Reflect modes across x-y plane (along z axis)

        See "Gravitational-wave modes from precessing black-hole binaries" by
        Boyle et al. (2014) for more details.

        """
        if self.dataType == UnknownDataType:
            raise ValueError("Cannot compute parity type for {0}.".format(self.data_type_string))
        W = self[:, :0]  # W without `data`, and `ells`=(0,-1); different from `self.copy_without_data()`
        W.data = np.empty_like(self.data)
        W.ells = self.ells
        W.frame = np.z_parity_conjugate(self.frame)
        s = self.spin_weight
        for ell in range(W.ell_min, W.ell_max + 1):
            if ((ell + s) % 2) == 0:
                lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1)]
                W.data[:, lm_indices] = np.conjugate(self.data[:, list(reversed(lm_indices))])
            else:
                lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1)]
                W.data[:, lm_indices] = -np.conjugate(self.data[:, list(reversed(lm_indices))])
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.z_parity_conjugate'.format(W, self))
        return W

    @property
    @waveform_alterations
    def z_parity_symmetric_part(self):
        """Component of waveform invariant under `z_parity_conjugate`"""
        W = self.z_parity_conjugate
        W.data = 0.5 * (self.data + W.data)
        W.frame = np.z_parity_symmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.z_parity_symmetric_part'.format(W, self))
        return W

    @property
    @waveform_alterations
    def z_parity_antisymmetric_part(self):
        """Component of waveform that changes sign under `z_parity_conjugate`"""
        W = self.z_parity_conjugate
        W.data = 0.5 * (self.data - W.data)
        W.frame = np.z_parity_antisymmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.z_parity_antisymmetric_part'.format(W, self))
        return W

    @property
    def z_parity_violation_squared(self):
        """(Squared) norm of z-parity-antisymmetric component of waveform"""
        return self.z_parity_antisymmetric_part.norm()

    @property
    def z_parity_violation_normalized(self):
        """Norm of z-parity-antisymmetric component divided by norm of waveform"""
        return np.sqrt(self.z_parity_antisymmetric_part.norm() / self.norm())

    @property
    @waveform_alterations
    def parity_conjugate(self):
        """Reflect modes along all axes

        See "Gravitational-wave modes from precessing black-hole binaries" by
        Boyle et al. (2014) for more details.

        """
        if self.dataType == UnknownDataType:
            raise ValueError("Cannot compute parity type for {0}.".format(self.data_type_string))
        W = self[:, :0]  # W without `data`, and `ells`=(0,-1); different from `self.copy_without_data()`
        W.data = np.empty_like(self.data)
        W.ells = self.ells
        W.frame = np.parity_conjugate(self.frame)
        s = self.spin_weight
        for ell in range(W.ell_min, W.ell_max + 1):
            lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1) if ((ell + s + m) % 2) == 0]
            W.data[:, lm_indices] = np.conjugate(self.data[:, list(reversed(lm_indices))])
            lm_indices = [sf.LM_index(ell, m, W.ell_min) for m in range(-ell, ell + 1) if ((ell + s + m) % 2) != 0]
            W.data[:, lm_indices] = -np.conjugate(self.data[:, list(reversed(lm_indices))])
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.parity_conjugate'.format(W, self))
        return W

    @property
    @waveform_alterations
    def parity_symmetric_part(self):
        """Component of waveform invariant under `parity_conjugate`"""
        W = self.parity_conjugate
        W.data = 0.5 * (self.data + W.data)
        W.frame = np.parity_symmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.parity_symmetric_part'.format(W, self))
        return W

    @property
    @waveform_alterations
    def parity_antisymmetric_part(self):
        """Component of waveform that changes sign under `parity_conjugate`"""
        W = self.parity_conjugate
        W.data = 0.5 * (self.data - W.data)
        W.frame = np.parity_antisymmetric_part(self.frame)
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.parity_antisymmetric_part'.format(W, self))
        return W

    @property
    def parity_violation_squared(self):
        """(Squared) norm of parity-antisymmetric component of waveform"""
        return self.parity_antisymmetric_part.norm()

    @property
    def parity_violation_normalized(self):
        """Norm of parity-antisymmetric component divided by norm of waveform"""
        return np.sqrt(self.parity_antisymmetric_part.norm() / self.norm())

    @waveform_alterations
    def copy_without_data(self):
        W = super(WaveformModes, self).copy_without_data()
        W.ells = 0, -1
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}.copy_without_data()'.format(W, self))
        return W

    @waveform_alterations
    def __getitem__(self, key):
        """Extract subsets of the data efficiently

        Note that if a second index is given to this function, it corresponds to modes, rather than data indices. See
        the docstring of the WaveformModes class for examples.

        """
        # Remove trivial tuple structure first
        if isinstance(key, tuple) and len(key) == 1:
            key = key[0]

        # Now figure out which type of return is desired
        if isinstance(key, tuple) and len(key) == 2:
            # Return a subset of the data from a subset of times
            if isinstance(key[1], int):
                if key[1] < self.ell_min or key[1] > self.ell_max:
                    raise ValueError("Requested ell value {0} lies outside ".format(key[1]) +
                                     "WaveformModes object's ell range ({0},{1}).".format(self.ell_min, self.ell_max))
                new_ell_min = key[1]
                new_ell_max = key[1]
                new_slice = slice(new_ell_min ** 2 - self.ell_min ** 2,
                                  new_ell_max * (new_ell_max + 2) + 1 - self.ell_min ** 2)
            elif isinstance(key[1], slice):
                if key[1].step and key[1].step != 1:
                    raise ValueError(
                        "Can only slice WaveformModes over contiguous ell values (step={0})".format(key[1].step))
                if not key[1].start and key[1].stop == 0:
                    new_ell_min = 0
                    new_ell_max = -1
                    new_slice = slice(0)
                else:
                    if not key[1].start:
                        new_ell_min = self.ell_min
                    else:
                        new_ell_min = key[1].start
                    if not key[1].stop:
                        new_ell_max = self.ell_max
                    else:
                        new_ell_max = key[1].stop - 1
                    if new_ell_min < self.ell_min or new_ell_max > self.ell_max:
                        raise ValueError(
                            "Requested ell range [{0},{1}] lies outside ".format(new_ell_min, new_ell_max) +
                            "WaveformBase's ell range [{0},{1}].".format(self.ell_min, self.ell_max))
                    new_slice = slice(new_ell_min ** 2 - self.ell_min ** 2,
                                      new_ell_max * (new_ell_max + 2) + 1 - self.ell_min ** 2)
            else:
                raise ValueError("Don't know what to do with slice of type `{0}`".format(type(key[1])))
            W = super(WaveformModes, self).__getitem__((key[0], new_slice))
            W.ells = new_ell_min, new_ell_max
        elif isinstance(key, slice) or isinstance(key, int):
            # Return complete data from a subset of times (key is slice), or
            # return complete data from a single instant in time (key is int)
            W = super(WaveformModes, self).__getitem__(key)
            W.ells = self.ells
        else:
            raise ValueError("Could not understand input `{0}` (of type `{1}`) ".format(key, type(key)))

        W.history.pop()  # remove WaveformBase history append, to be replaced next
        W.__history_depth__ -= 1
        W._append_history('{0} = {1}[{2}]'.format(W, self, key))

        return W

    def __repr__(self):
        # "The goal of __str__ is to be readable; the goal of __repr__ is to be unambiguous." --- stackoverflow
        rep = super(WaveformModes, self).__repr__()
        rep += "\n# ell_min={0}, ell_max={1}".format(self.ell_min, self.ell_max)
        return rep


    # N.B.: There are additional methods attached to `WaveformModes` in the `waveform_base` file.  These functions
    # cannot be added here, because they depend on `WaveformGrid` objects, which depend on `WaveformModes` objects.



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import

from . import (Inertial, WaveformModes, SpinWeights, h, sigma, psi0, psi1, psi2, psi3)
from .waveform_base import WaveformBase, waveform_alterations

import sys
import warnings
import numbers
import math
import numpy as np
from scipy import interpolate
import quaternion
import spherical_functions as sf
import spinsfast


def process_transformation_kwargs(**kwargs):
    original_kwargs = kwargs.copy()

    # Build the supertranslation and spacetime_translation arrays
    supertranslation = np.zeros((4,), dtype=complex)  # For now; may be resized below
    ell_max_supertranslation = 1  # For now; may be increased below
    if 'supertranslation' in kwargs:
        supertranslation = np.array(kwargs.pop('supertranslation'), dtype=complex)
        if supertranslation.dtype != 'complex' and supertranslation.size > 0:
            # I don't actually think this can ever happen...
            raise TypeError("\nInput argument `supertranslation` should be a complex array with size>0.\n"
                            "Got a {0} array of shape {1}.".format(supertranslation.dtype,
                                                                   supertranslation.shape))
        # Make sure the array has size at least 4, by padding with zeros
        if supertranslation.size <= 4:
            supertranslation = np.lib.pad(supertranslation, (0, 4-supertranslation.size),
                                          'constant', constant_values=(0.0,))
        # Check that the shape is a possible array of scalar modes with complete (ell,m) data
        ell_max_supertranslation = int(np.sqrt(len(supertranslation))) - 1
        if (ell_max_supertranslation + 1)**2 != len(supertranslation):
            raise ValueError('\nInput supertranslation parameter must contain modes from ell=0 up to some ell_max, '
                             'including\nall relevant m modes in standard order (see `spherical_functions` '
                             'documentation for details).\nThus, it must be an array with length given by a '
                             'perfect square; its length is {0}'.format(len(supertranslation)))
        # Check that the resulting supertranslation will be real
        for ell in range(ell_max_supertranslation+1):
            for m in range(ell+1):
                i_pos = sf.LM_index(ell, m, 0)
                i_neg = sf.LM_index(ell, -m, 0)
                a = supertranslation[i_pos]
                b = supertranslation[i_neg]
                if abs(a - (-1.)**m * b.conjugate()) > 3e-16 + 1e-15 * abs(b):
                    raise ValueError("\nsupertranslation[{0}]={1}  # (ell,m)=({2},{3})\n".format(i_pos, a, ell, m)
                                     + "supertranslation[{0}]={1}  # (ell,m)=({2},{3})\n".format(i_neg, b, ell, -m)
                                     + "Will result in an imaginary supertranslation.")
    spacetime_translation = np.zeros((4,), dtype=float)
    spacetime_translation[0] = sf.constant_from_ell_0_mode(supertranslation[0]).real
    spacetime_translation[1:4] = -sf.vector_from_ell_1_modes(supertranslation[1:4]).real
    if 'spacetime_translation' in kwargs:
        st_trans = np.array(kwargs.pop('spacetime_translation'), dtype=float)
        if st_trans.shape != (4,) or st_trans.dtype != 'float':
            raise TypeError("\nInput argument `spacetime_translation` should be a float array of shape (4,).\n"
                            "Got a {0} array of shape {1}.".format(st_trans.dtype, st_trans.shape))
        spacetime_translation = st_trans[:]
        supertranslation[0] = sf.constant_as_ell_0_mode(spacetime_translation[0])
        supertranslation[1:4] = sf.vector_as_ell_1_modes(-spacetime_translation[1:4])
    if 'space_translation' in kwargs:
        s_trans = np.array(kwargs.pop('space_translation'), dtype=float)
        if s_trans.shape != (3,) or s_trans.dtype != 'float':
            raise TypeError("\nInput argument `space_translation` should be an array of floats of shape (3,).\n"
                            "Got a {0} array of shape {1}.".format(s_trans.dtype, s_trans.shape))
        spacetime_translation[1:4] = s_trans[:]
        supertranslation[1:4] = sf.vector_as_ell_1_modes(-spacetime_translation[1:4])
    if 'time_translation' in kwargs:
        t_trans = kwargs.pop('time_translation')
        if not isinstance(t_trans, float):
            raise TypeError("\nInput argument `time_translation` should be a single float.\n"
                            "Got {0}.".format(t_trans))
        spacetime_translation[0] = t_trans
        supertranslation[0] = sf.constant_as_ell_0_mode(spacetime_translation[0])

    # Decide on the number of points to use in each direction.  A nontrivial supertranslation will introduce
    # power in higher modes, so for best accuracy, we need to account for that.  But we'll make it a firm
    # requirement to have enough points to capture the original waveform, at least
    w_ell_max = kwargs.pop('ell_max')
    ell_max = w_ell_max + ell_max_supertranslation
    n_theta = kwargs.pop('n_theta', 2*ell_max+1)
    n_phi = kwargs.pop('n_phi', 2*ell_max+1)
    if n_theta < 2*ell_max+1 and abs(supertranslation[1:]).max() > 0.0:
        warning = ("n_theta={0} is small; because of the supertranslation, ".format(n_theta)
                   + "it will lose accuracy for anything less than 2*ell+1={1}".format(ell_max))
        warnings.warn(warning)
    if n_theta < 2*w_ell_max+1:
        raise ValueError('n_theta={0} is too small; '.format(n_theta)
                         + 'must be at least 2*ell+1={1}'.format(2*w_ell_max+1))
    if n_phi < 2*ell_max+1 and abs(supertranslation[1:]).max() > 0.0:
        warning = ("n_phi={0} is small; because of the supertranslation, ".format(n_phi)
                   + "it will lose accuracy for anything less than 2*ell+1={1}".format(ell_max))
        warnings.warn(warning)
    if n_phi < 2*w_ell_max+1:
        raise ValueError('n_phi={0} is too small; '.format(n_phi)
                         + 'must be at least 2*ell+1={1}'.format(2*w_ell_max+1))

    # Get the rotor for the frame rotation
    frame_rotation = np.quaternion(*np.array(kwargs.pop('frame_rotation', [1, 0, 0, 0]), dtype=float))
    if frame_rotation.abs() < 3e-16:
        raise ValueError('frame_rotation={0} should be a unit quaternion'.format(frame_rotation))
    frame_rotation = frame_rotation.normalized()

    # Get the boost velocity vector
    boost_velocity = np.array(kwargs.pop('boost_velocity', [0.0]*3), dtype=float)
    beta = np.linalg.norm(boost_velocity)
    if boost_velocity.shape != (3,) or beta >= 1.0:
        raise ValueError('Input boost_velocity=`{0}` should be a 3-vector with '
                         'magnitude strictly less than 1.0.'.format(boost_velocity))
    gamma = 1 / math.sqrt(1 - beta**2)
    varphi = math.atanh(beta)

    if kwargs:
        import pprint
        warnings.warn("\nUnused kwargs passed to this function:\n{0}".format(pprint.pformat(kwargs, width=1)))

    # These are the angles in the transformed system at which we need to know the function values
    thetaprm_j_phiprm_k = np.array([[[thetaprm_j, phiprm_k]
                                     for phiprm_k in np.linspace(0.0, 2*np.pi, num=n_phi, endpoint=False)]
                                    for thetaprm_j in np.linspace(0.0, np.pi, num=n_theta, endpoint=True)])

    # Construct the function that modifies our rotor grid to account for the boost
    if beta > 3e-14:  # Tolerance for beta; any smaller and numerical errors will have greater effect
        vhat = boost_velocity / beta

        def Bprm_j_k(thetaprm, phiprm):
            """Construct rotor taking r' to r

            I derived this result in a different way, but I've also found it described in Penrose-Rindler Vol. 1,
            around Eq. (1.3.5).  Note, however, that their discussion is for the past celestial sphere,
            so there's a sign difference.

            """
            # Note: It doesn't matter which we use -- r' or r; all we need is the direction of the bivector
            # spanned by v and r', which is the same as the direction of the bivector spanned by v and r,
            # since either will be normalized, and one cross product is zero iff the other is zero.
            rprm = np.array([math.cos(phiprm)*math.sin(thetaprm),
                             math.sin(phiprm)*math.sin(thetaprm),
                             math.cos(thetaprm)])
            Thetaprm = math.acos(np.dot(vhat, rprm))
            Theta = 2 * math.atan(math.exp(-varphi) * math.tan(Thetaprm/2.0))
            rprm_cross_vhat = np.quaternion(0.0, *np.cross(rprm, vhat))
            if rprm_cross_vhat.abs() > 1e-200:
                return (rprm_cross_vhat.normalized() * (Thetaprm - Theta) / 2).exp()
            else:
                return quaternion.one
    else:
        def Bprm_j_k(thetaprm, phiprm):
            return quaternion.one

    # Set up rotors that we can use to evaluate the SWSHs in the original frame
    R_j_k = np.empty(thetaprm_j_phiprm_k.shape[:2], dtype=np.quaternion)
    for j in range(thetaprm_j_phiprm_k.shape[0]):
        for k in range(thetaprm_j_phiprm_k.shape[1]):
            thetaprm_j, phiprm_k = thetaprm_j_phiprm_k[j, k]
            R_j_k[j, k] = (Bprm_j_k(thetaprm_j, phiprm_k)
                           * frame_rotation
                           * quaternion.from_spherical_coords(thetaprm_j, phiprm_k))

    return (
        supertranslation, ell_max, n_theta, n_phi,
        beta, gamma, varhpi,
        R_j_k, Bprm_j_k, thetaprm_j_phiprm_k,
    )


class WaveformGrid(WaveformBase):

    def __init__(self, *args, **kwargs):
        """Initializer for WaveformGrid object"""
        # Do not directly access __n_theta or __n_phi; use n_theta or n_phi instead
        self.__n_theta = kwargs.pop('n_theta', 0)
        self.__n_phi = kwargs.pop('n_phi', 0)
        super(WaveformGrid, self).__init__(*args, **kwargs)

    @waveform_alterations
    def ensure_validity(self, alter=True, assertions=False):
        """Try to ensure that the `WaveformGrid` object is valid

        See `WaveformBase.ensure_validity` for the basic tests.  This function also includes tests that `data` is
        complex, and consistent with the n_theta and n_phi values.

        """
        import numbers

        errors = []
        alterations = []

        if assertions:
            from .waveform_base import test_with_assertions
            test = test_with_assertions
        else:
            from .waveform_base import test_without_assertions
            test = test_without_assertions

        test(errors,
             isinstance(self.__n_theta, numbers.Integral),
             'isinstance(self.__n_theta, numbers.Integral)  # type(self.__n_theta)={0}'.format(type(self.__n_theta)))
        test(errors,
             isinstance(self.__n_phi, numbers.Integral),
             'isinstance(self.__n_phi, numbers.Integral)  # type(self.__n_phi)={0}'.format(type(self.__n_phi)))
        test(errors,
             self.__n_theta >= 0,
             'self.__n_theta>=0 # {0}'.format(self.__n_theta))
        test(errors,
             self.__n_phi >= 0,
             'self.__n_phi>=0 # {0}'.format(self.__n_phi))

        test(errors,
             self.data.dtype == np.dtype(np.complex),
             'self.data.dtype == np.dtype(np.complex)  # self.data.dtype={0}'.format(self.data.dtype))
        test(errors,
             self.data.ndim >= 2,
             'self.data.ndim >= 2 # self.data.ndim={0}'.format(self.data.ndim))
        test(errors,
             self.data.shape[1] == self.__n_theta * self.__n_phi,
             'self.data.shape[1] == self.__n_theta * self.__n_phi  '
             '# self.data.shape={0}; self.__n_theta * self.__n_phi={1}'.format(self.data.shape[1],
                                                                               self.__n_theta * self.__n_phi))

        if alterations:
            self._append_history(alterations)
            print("The following alterations were made:\n\t" + '\n\t'.join(alterations))
        if errors:
            print("The following conditions were found to be incorrectly False:\n\t" + '\n\t'.join(errors))
            return False

        # Call the base class's version
        super(WaveformGrid, self).ensure_validity(alter, assertions)

        self.__history_depth__ -= 1
        self._append_history('WaveformModes.ensure_validity' +
                             '({0}, alter={1}, assertions={2})'.format(self, alter, assertions))

        return True

    @property
    def n_theta(self):
        return self.__n_theta

    @property
    def n_phi(self):
        return self.__n_phi

    def to_modes(self, ell_max=None):
        """Transform to modes of a spin-weighted spherical harmonic expansion

        Parameters
        ----------
        self : WaveformGrid object
            This is the object to be transformed to SWSH modes
        ell_max : int, optional
            The highest ell value to include in the output data.  Default value is deduced from n_theta and n_phi.

        """
        s = SpinWeights[self.dataType]
        ell_min = abs(s)
        if ell_max is None:
            ell_max = int((max(self.n_theta, self.n_phi)-1)//2)
        if not isinstance(ell_max, numbers.Integral) or ell_max < 0:
            raise ValueError("Input `ell_max` should be a nonnegative integer; got `{0}` instead".format(ell_max))

        final_dim = int(np.prod(self.data.shape[2:]))
        old_data = self.data.reshape((self.n_times, self.n_theta, self.n_phi, final_dim))
        new_data = np.empty((self.n_times, sf.LM_total_size(ell_min, ell_max), final_dim), dtype=complex)
        # Note that spinsfast returns all modes, including ell<abs(s).  So we just chop those off
        for i_time in range(self.n_times):
            for i_final in range(final_dim):
                new_data[i_time, :, i_final] = spinsfast.map2salm(old_data[i_time, :, :, i_final], s, ell_max)\
                    [sf.LM_index(ell_min, -ell_min, 0):]
        new_data = new_data.reshape((self.n_times, sf.LM_total_size(ell_min, ell_max))+self.data.shape[2:])

        # old_data = self.data.reshape((self.n_times, self.n_theta, self.n_phi)+self.data.shape[2:])
        # new_data = np.empty((self.n_times, sf.LM_total_size(ell_min, ell_max))+self.data.shape[2:], dtype=complex)
        # # Note that spinsfast returns all modes, including ell<abs(s).  So we just chop those off
        # for i_time in range(self.n_times):
        #     new_data[i_time, :] = spinsfast.map2salm(old_data[i_time, :, :], s, ell_max)\
        #         [sf.LM_index(ell_min, -ell_min, 0):]

        m = WaveformModes(t=self.t, data=new_data, history=self.history,
                          ell_min=ell_min, ell_max=ell_max,
                          frameType=self.frameType, dataType=self.dataType,
                          r_is_scaled_out=self.r_is_scaled_out, m_is_scaled_out=self.m_is_scaled_out,
                          constructor_statement="{0}.to_modes({1})".format(self, ell_max))
        return m

    @classmethod
    def from_modes(cls, w_modes, **kwargs):
        """Construct grid object from modes, with optional BMS transformation

        This "constructor" is designed with the goal of transforming the frame in which the modes are measured.  If
        this is not desired, it can be called without those parameters.

        It is important to note that the input transformation parameters are applied in the order listed in the
        parameter list below:
          1. (Super)Translations
          2. Rotation (about the origin of the translated system)
          3. Boost
        All input parameters refer to the transformation required to take the mode's inertial frame onto the inertial
        frame of the grid's inertial observers.  In what follows, the inertial frame of the modes will be unprimed,
        while the inertial frame of the grid will be primed.

        The translations (space, time, spacetime, or super) can be given in various ways, which may override each
        other.  Ultimately, however, they are essentially combined into a single function `alpha`, representing the
        supertranslation, which transforms the asymptotic time variable `u` as
          u'(theta, phi) = u - alpha(theta, phi)
        A simple time translation would correspond to
          alpha(theta, phi) = time_translation
        A pure spatial translation would correspond to
          alpha(theta, phi) = np.dot(space_translation, -nhat(theta, phi))
        where `np.dot` is the usual dot product, and `nhat` is the unit vector in the given direction.


        Parameters
        ----------
        w_modes : WaveformModes
            The object storing the modes of the original waveform, which will be converted to values on a grid in
            this function.  This is the only required argument to this function.
        n_theta : int, optional
        n_phi : int, optional
            Number of points in the equi-angular grid in the colatitude (theta) and azimuth (phi) directions. Each
            defaults to 2*ell_max+1, which is optimal for accuracy and speed.  However, this ell_max will typically
            be greater than the input waveform's ell_max by at least one, or the ell_max of the input
            supertranslation (whichever is greater).  This is to minimally account for the power at higher orders
            that such a supertranslation introduces.  You may wish to increase this further if the spatial size of
            your supertranslation is large compared to the smallest wavelength you wish to capture in your data
            [e.g., ell_max*Omega_orbital_max/speed_of_light], or if your boost speed is close to the speed of light.
        time_translation : float, optional
            Defaults to zero.  Nonzero overrides spacetime_translation and supertranslation.
        space_translation : float array of length 3, optional
            Defaults to empty (no translation).  Non-empty overrides spacetime_translation and supertranslation.
        spacetime_translation : float array of length 4, optional
            Defaults to empty (no translation).  Non-empty overrides supertranslation.
        supertranslation : complex array, optional
            This gives the complex components of the spherical-harmonic expansion of the supertranslation in standard
            form, starting from ell=0 up to some ell_max, which may be different from the ell_max of the input
            WaveformModes object.  Supertranslations must be real, so these values must obey the condition
              alpha^{ell,m} = (-1)^m \bar{alpha}^{ell,-m}
            Defaults to empty, which causes no supertranslation.
        frame_rotation : quaternion, optional
            Transformation applied to (x,y,z) basis of the mode's inertial frame.  For example, the basis z vector of
            the new grid frame may be written as
              z' = frame_rotation * z * frame_rotation.inverse()
            Defaults to 1, corresponding to the identity transformation (no frame_rotation).
        boost_velocity : float array of length 3, optional
            This is the three-velocity vector of the grid frame relative to the mode frame.  The norm of this vector
            is checked to ensure that it is smaller than 1.  Defaults to [], corresponding to no boost.

        Returns
        -------
        WaveformGrid

        """
        # Check input object type and frame type
        #
        # The data in `w_modes` is given in the original frame.  We need to get the value of the field on a grid of
        # points corresponding to the points in the new grid frame.  But we must also remember that this is a
        # spin-weighted and boost-weighted field, which means that we need to account for the frame_rotation due to
        # `frame_rotation` and `boost_velocity`.  The way to do this is to use rotors to transform the points as needed,
        # and evaluate the SWSHs.  But for now, let's just reject any waveforms in a non-inertial frame
        if not isinstance(w_modes, WaveformModes):
            raise TypeError("\nInput waveform object must be an instance of `WaveformModes`; "
                            "this is of type `{0}`".format(type(w_modes).__name__))
        if w_modes.frameType != Inertial:
            raise ValueError("\nInput waveform object must be in an inertial frame; "
                             "this is in a frame of type `{0}`".format(w_modes.frame_type_string))

        # The first task is to establish a set of constant u' slices on which the new grid should be evaluated.  This
        # is done simply by translating the original set of slices by the time translation (the lowest moment of the
        # supertranslation).  But some of these slices (at the beginning and end) will not have complete data,
        # because of the direction-dependence of the rest of the supertranslation.  That is, in some directions,
        # the data for the complete slice (data in all directions on the sphere) of constant u' will actually refer to
        # spacetime events that were not in the original set of time slices; we would have to extrapolate the original
        # data.  So, for nontrivial supertranslations, the output data will actually represent a proper subset of the
        # input data.
        #
        # We can invert the equation for u' to obtain u as a function of angle assuming constant u'
        #   u'(theta, phi) = u + alpha(theta, phi) + u * np.dot(boost_velocity, nhat(theta, phi))
        #   u(theta, phi) = (u' - alpha(theta, phi)) / (1 + np.dot(boost_velocity, nhat(theta, phi)))
        # But really, we want u'(theta', phi') for given values
        #
        # Note that `space_translation` (and the spatial part of `spacetime_translation`) get reversed signs when
        # transformed into supertranslation modes, because these pieces enter the time transformation with opposite
        # sign compared to the time translation, as can be seen by looking at the retarded time: `t-r`.

        original_kwargs = kwargs.copy()

        # Build the supertranslation and spacetime_translation arrays
        supertranslation = np.zeros((4,), dtype=complex)  # For now; may be resized below
        ell_max_supertranslation = 1  # For now; may be increased below
        if 'supertranslation' in kwargs:
            supertranslation = np.array(kwargs.pop('supertranslation'), dtype=complex)
            if supertranslation.dtype != 'complex' and supertranslation.size > 0:
                # I don't actually think this can ever happen...
                raise TypeError("\nInput argument `supertranslation` should be a complex array with size>0.\n"
                                "Got a {0} array of shape {1}.".format(supertranslation.dtype,
                                                                       supertranslation.shape))
            # Make sure the array has size at least 4, by padding with zeros
            if supertranslation.size <= 4:
                supertranslation = np.lib.pad(supertranslation, (0, 4-supertranslation.size),
                                              'constant', constant_values=(0.0,))
            # Check that the shape is a possible array of scalar modes with complete (ell,m) data
            ell_max_supertranslation = int(np.sqrt(len(supertranslation))) - 1
            if (ell_max_supertranslation + 1)**2 != len(supertranslation):
                raise ValueError('\nInput supertranslation parameter must contain modes from ell=0 up to some ell_max, '
                                 'including\nall relevant m modes in standard order (see `spherical_functions` '
                                 'documentation for details).\nThus, it must be an array with length given by a '
                                 'perfect square; its length is {0}'.format(len(supertranslation)))
            # Check that the resulting supertranslation will be real
            for ell in range(ell_max_supertranslation+1):
                for m in range(ell+1):
                    i_pos = sf.LM_index(ell, m, 0)
                    i_neg = sf.LM_index(ell, -m, 0)
                    a = supertranslation[i_pos]
                    b = supertranslation[i_neg]
                    if abs(a - (-1.)**m * b.conjugate()) > 3e-16 + 1e-15 * abs(b):
                        raise ValueError("\nsupertranslation[{0}]={1}  # (ell,m)=({2},{3})\n".format(i_pos, a, ell, m)
                                         + "supertranslation[{0}]={1}  # (ell,m)=({2},{3})\n".format(i_neg, b, ell, -m)
                                         + "Will result in an imaginary supertranslation.")
        spacetime_translation = np.zeros((4,), dtype=float)
        spacetime_translation[0] = sf.constant_from_ell_0_mode(supertranslation[0]).real
        spacetime_translation[1:4] = -sf.vector_from_ell_1_modes(supertranslation[1:4]).real
        if 'spacetime_translation' in kwargs:
            st_trans = np.array(kwargs.pop('spacetime_translation'), dtype=float)
            if st_trans.shape != (4,) or st_trans.dtype != 'float':
                raise TypeError("\nInput argument `spacetime_translation` should be a float array of shape (4,).\n"
                                "Got a {0} array of shape {1}.".format(st_trans.dtype, st_trans.shape))
            spacetime_translation = st_trans[:]
            supertranslation[0] = sf.constant_as_ell_0_mode(spacetime_translation[0])
            supertranslation[1:4] = sf.vector_as_ell_1_modes(-spacetime_translation[1:4])
        if 'space_translation' in kwargs:
            s_trans = np.array(kwargs.pop('space_translation'), dtype=float)
            if s_trans.shape != (3,) or s_trans.dtype != 'float':
                raise TypeError("\nInput argument `space_translation` should be an array of floats of shape (3,).\n"
                                "Got a {0} array of shape {1}.".format(s_trans.dtype, s_trans.shape))
            spacetime_translation[1:4] = s_trans[:]
            supertranslation[1:4] = sf.vector_as_ell_1_modes(-spacetime_translation[1:4])
        if 'time_translation' in kwargs:
            t_trans = kwargs.pop('time_translation')
            if not isinstance(t_trans, float):
                raise TypeError("\nInput argument `time_translation` should be a single float.\n"
                                "Got {0}.".format(t_trans))
            spacetime_translation[0] = t_trans
            supertranslation[0] = sf.constant_as_ell_0_mode(spacetime_translation[0])

        # Decide on the number of points to use in each direction.  A nontrivial supertranslation will introduce
        # power in higher modes, so for best accuracy, we need to account for that.  But we'll make it a firm
        # requirement to have enough points to capture the original waveform, at least
        ell_max = w_modes.ell_max + ell_max_supertranslation
        n_theta = kwargs.pop('n_theta', 2*ell_max+1)
        n_phi = kwargs.pop('n_phi', 2*ell_max+1)
        if n_theta < 2*ell_max+1 and abs(supertranslation[1:]).max() > 0.0:
            warning = ("n_theta={0} is small; because of the supertranslation, ".format(n_theta)
                       + "it will lose accuracy for anything less than 2*ell+1={1}".format(ell_max))
            warnings.warn(warning)
        if n_theta < 2*w_modes.ell_max+1:
            raise ValueError('n_theta={0} is too small; '.format(n_theta)
                             + 'must be at least 2*ell+1={1}'.format(2*w_modes.ell_max+1))
        if n_phi < 2*ell_max+1 and abs(supertranslation[1:]).max() > 0.0:
            warning = ("n_phi={0} is small; because of the supertranslation, ".format(n_phi)
                       + "it will lose accuracy for anything less than 2*ell+1={1}".format(ell_max))
            warnings.warn(warning)
        if n_phi < 2*w_modes.ell_max+1:
            raise ValueError('n_phi={0} is too small; '.format(n_phi)
                             + 'must be at least 2*ell+1={1}'.format(2*w_modes.ell_max+1))

        # Get the rotor for the frame rotation
        frame_rotation = np.quaternion(*np.array(kwargs.pop('frame_rotation', [1, 0, 0, 0]), dtype=float))
        if frame_rotation.abs() < 3e-16:
            raise ValueError('frame_rotation={0} should be a unit quaternion'.format(frame_rotation))
        frame_rotation = frame_rotation.normalized()

        # Get the boost velocity vector
        boost_velocity = np.array(kwargs.pop('boost_velocity', [0.0]*3), dtype=float)
        beta = np.linalg.norm(boost_velocity)
        if boost_velocity.shape != (3,) or beta >= 1.0:
            raise ValueError('Input boost_velocity=`{0}` should be a 3-vector with '
                             'magnitude strictly less than 1.0.'.format(boost_velocity))
        gamma = 1 / math.sqrt(1 - beta**2)
        varphi = math.atanh(beta)

        if kwargs:
            import pprint
            warnings.warn("\nUnused kwargs passed to this function:\n{0}".format(pprint.pformat(kwargs, width=1)))

        # These are the angles in the transformed system at which we need to know the function values
        thetaprm_j_phiprm_k = np.array([[[thetaprm_j, phiprm_k]
                                         for phiprm_k in np.linspace(0.0, 2*np.pi, num=n_phi, endpoint=False)]
                                        for thetaprm_j in np.linspace(0.0, np.pi, num=n_theta, endpoint=True)])

        # Construct the function that modifies our rotor grid to account for the boost
        if beta > 3e-14:  # Tolerance for beta; any smaller and numerical errors will have greater effect
            vhat = boost_velocity / beta

            def Bprm_j_k(thetaprm, phiprm):
                """Construct rotor taking r' to r

                I derived this result in a different way, but I've also found it described in Penrose-Rindler Vol. 1,
                around Eq. (1.3.5).  Note, however, that their discussion is for the past celestial sphere,
                so there's a sign difference.

                """
                # Note: It doesn't matter which we use -- r' or r; all we need is the direction of the bivector
                # spanned by v and r', which is the same as the direction of the bivector spanned by v and r,
                # since either will be normalized, and one cross product is zero iff the other is zero.
                rprm = np.array([math.cos(phiprm)*math.sin(thetaprm),
                                 math.sin(phiprm)*math.sin(thetaprm),
                                 math.cos(thetaprm)])
                Thetaprm = math.acos(np.dot(vhat, rprm))
                Theta = 2 * math.atan(math.exp(-varphi) * math.tan(Thetaprm/2.0))
                rprm_cross_vhat = np.quaternion(0.0, *np.cross(rprm, vhat))
                if rprm_cross_vhat.abs() > 1e-200:
                    return (rprm_cross_vhat.normalized() * (Thetaprm - Theta) / 2).exp()
                else:
                    return quaternion.one
        else:
            def Bprm_j_k(thetaprm, phiprm):
                return quaternion.one

        # Set up rotors that we can use to evaluate the SWSHs in the original frame
        R_j_k = np.empty(thetaprm_j_phiprm_k.shape[:2], dtype=np.quaternion)
        for j in range(thetaprm_j_phiprm_k.shape[0]):
            for k in range(thetaprm_j_phiprm_k.shape[1]):
                thetaprm_j, phiprm_k = thetaprm_j_phiprm_k[j, k]
                R_j_k[j, k] = (Bprm_j_k(thetaprm_j, phiprm_k)
                               * frame_rotation
                               * quaternion.from_spherical_coords(thetaprm_j, phiprm_k))

        # TODO: Incorporate the w_modes.frame information into rotors, which will require time dependence throughout
        # It would be best to leave the waveform in its frame.  But we'll have to apply the frame_rotation to the BMS
        # elements, which will be a little tricky.  Still probably not as tricky as applying to the waveform...

        # We need values of (1) waveform, (2) conformal factor, and (3) supertranslation, at each point of the
        # transformed grid, at each instant of time.
        SWSH_j_k = sf.SWSH_grid(R_j_k, w_modes.spin_weight, ell_max)
        SH_j_k = sf.SWSH_grid(R_j_k, 0, ell_max_supertranslation)  # standard (spin-zero) spherical harmonics
        r_j_k = np.array([(R*quaternion.z*R.inverse()).vec for R in R_j_k.flat]).T
        kconformal_j_k = 1. / (gamma*(1-np.dot(boost_velocity, r_j_k).reshape(R_j_k.shape)))
        alphasupertranslation_j_k = np.tensordot(supertranslation, SH_j_k, axes=([0], [2])).real
        fprm_i_j_k = np.tensordot(
            w_modes.data, SWSH_j_k[:, :, sf.LM_index(w_modes.ell_min, -w_modes.ell_min, 0)
                                         :sf.LM_index(w_modes.ell_max, w_modes.ell_max, 0)+1],
            axes=([1], [2]))
        if w_modes.dataType == h:
            # Note that SWSH_j_k will use s=-2 in this case, so it can be used in the tensordot correctly
            supertranslation_deriv = sf.ethbar_GHP(sf.ethbar_GHP(supertranslation, 0, 0), -1, 0)
            supertranslation_deriv_values = np.tensordot(
                supertranslation_deriv,
                SWSH_j_k[:, :, :sf.LM_index(ell_max_supertranslation, ell_max_supertranslation, 0)+1],
                axes=([0], [2]))
            fprm_i_j_k -= supertranslation_deriv_values[np.newaxis, :, :]
        elif w_modes.dataType == sigma:
            # Note that SWSH_j_k will use s=+2 in this case, so it can be used in the tensordot correctly
            supertranslation_deriv = sf.eth_GHP(sf.eth_GHP(supertranslation, 0, 0), 1, 0)
            supertranslation_deriv_values = np.tensordot(
                supertranslation_deriv,
                SWSH_j_k[:, :, :sf.LM_index(ell_max_supertranslation, ell_max_supertranslation, 0)+1],
                axes=([0], [2]))
            fprm_i_j_k -= supertranslation_deriv_values[np.newaxis, :, :]
        elif w_modes.dataType in [psi0, psi1, psi2, psi3]:
            warning = ("\nTechnically, waveforms of dataType `{0}` ".format(w_modes.data_type_string)
                       + "do not transform among themselves;\n there is mixing from psi_n, for each n greater than "
                       + "this waveform's.\nProceeding on the assumption other contributions are small.  However,\n"
                       + "note that it is possible to construct a `psin` data type containing all necessary modes.")
            warnings.warn(warning)

        fprm_i_j_k *= (gamma**w_modes.gamma_weight
                       * kconformal_j_k**w_modes.conformal_weight)[np.newaxis, :, :]

        # Determine the new time slices.  The set u' is chosen so that on each slice of constant u'_i, the average value
        # of u is precisely u_i.  But then, we have to narrow that set down, so that every physical point on all the
        # u'_i' slices correspond to data in the range of input data.
        uprm_i = (1/gamma) * (w_modes.t - spacetime_translation[0])
        uprm_min = (kconformal_j_k * (w_modes.t[0] - alphasupertranslation_j_k)).max()
        uprm_max = (kconformal_j_k * (w_modes.t[-1] - alphasupertranslation_j_k)).min()
        uprm_iprm = uprm_i[(uprm_i >= uprm_min) & (uprm_i <= uprm_max)]

        # Interpolate along each grid line to the new time in that direction.  Note that if there are additional
        # dimensions in the waveform data, InterpolatedUnivariateSpline will not be able to handle them automatically,
        # so we have to loop over them explicitly; an Ellipsis can't handle them.  Also, we are doing all time steps in
        # one go, for each j,k,... value, which means that we can overwrite the original data
        final_dim = int(np.prod(fprm_i_j_k.shape[3:]))
        fprm_i_j_k = fprm_i_j_k.reshape(fprm_i_j_k.shape[:3] + (final_dim,))
        for j in range(n_theta):
            for k in range(n_phi):
                uprm_i_j_k = kconformal_j_k[j, k] * (w_modes.t - alphasupertranslation_j_k[j, k])
                for final_indices in range(final_dim):
                    re_fprm_iprm_j_k = interpolate.InterpolatedUnivariateSpline(uprm_i_j_k,
                        fprm_i_j_k[:, j, k, final_indices].real)
                    im_fprm_iprm_j_k = interpolate.InterpolatedUnivariateSpline(uprm_i_j_k,
                        fprm_i_j_k[:, j, k, final_indices].imag)
                    fprm_i_j_k[:len(uprm_iprm), j, k, final_indices] = (
                        re_fprm_iprm_j_k(uprm_iprm) + 1j * im_fprm_iprm_j_k(uprm_iprm))

        # Delete the extra rows from fprm_i_j_k, corresponding to values of u' outside of [u'min, u'max]
        fprm_iprm_j_k = np.delete(fprm_i_j_k, np.s_[len(uprm_iprm):], 0)

        # Reshape, to have correct final dimensions
        fprm_iprm_j_k = fprm_iprm_j_k.reshape((fprm_iprm_j_k.shape[0], n_theta*n_phi)+w_modes.data.shape[2:])

        # Encapsulate into a new grid waveform
        g = cls(t=uprm_iprm, data=fprm_iprm_j_k, history=w_modes.history,
                n_theta=n_theta, n_phi=n_phi,
                frameType=w_modes.frameType, dataType=w_modes.dataType,
                r_is_scaled_out=w_modes.r_is_scaled_out, m_is_scaled_out=w_modes.m_is_scaled_out,
                constructor_statement="{0}.from_modes({1}, **{2})".format(cls.__name__, w_modes, original_kwargs))
        return g

    @classmethod
    def transform(cls, w_modes, **kwargs):
        """Transform modes by some BMS transformation

        This simply applies the `WaveformGrid.from_modes` function, followed by the `WaveformGrid.to_modes` function.
        See their respective docstrings for more details.  However, note that the `ell_max` parameter used in the
        second function call defaults here to the `ell_max` value in the input waveform.  This is slightly different
        from the usual default, because `WaveformGrid.from_modes` usually increases the effective ell value by 1.

        """
        if not isinstance(w_modes, WaveformModes):
            raise TypeError("Expected WaveformModes object in argument 1; "
                            "got `{0}` instead.".format(type(w_modes).__name__))
        ell_max = kwargs.pop('ell_max', w_modes.ell_max)
        return WaveformGrid.from_modes(w_modes, **kwargs).to_modes(ell_max)

    def __repr__(self):
        # "The goal of __str__ is to be readable; the goal of __repr__ is to be unambiguous." --- stackoverflow
        rep = super(WaveformGrid, self).__repr__()
        rep += "\n# n_theta={0}, n_phi={1}".format(self.n_theta, self.n_phi)
        return rep


# Now, we can assign WaveformModes objects new capabilities based on WaveformGrid functions
WaveformModes.to_grid = lambda w_modes, **kwargs: WaveformGrid.from_modes(w_modes, **kwargs)
WaveformModes.from_grid = classmethod(lambda cls, w_grid, ell_max: WaveformGrid.to_modes(w_grid, ell_max))
WaveformModes.transform = lambda w_mode, **kwargs: WaveformGrid.transform(w_mode, **kwargs)
if sys.version_info[0] == 2:
    WaveformModes.to_grid.__func__.__doc__ = WaveformGrid.from_modes.__doc__
    WaveformModes.from_grid.__func__.__doc__ = WaveformGrid.to_modes.__doc__
    WaveformModes.transform.__func__.__doc__ = WaveformGrid.transform.__doc__
else:
    WaveformModes.to_grid.__doc__ = WaveformGrid.from_modes.__doc__
    WaveformModes.from_grid.__func__.__doc__ = WaveformGrid.to_modes.__doc__
    WaveformModes.transform.__doc__ = WaveformGrid.transform.__doc__



# Copyright (c) 2018, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/scri/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import

from . import (Inertial, WaveformModes, SpinWeights, h, sigma, psi0, psi1, psi2, psi3)
from .waveform_base import WaveformBase, waveform_alterations


# class WaveformCollection(object):
#     """A collection of waveforms to be treated uniformly

#     This is a simple container that collects several Waveform objects.  They are all required to 

#     """
#     def __init__(self, *args, **kwargs):
        



"""

Read output with

    import pandas as pd
    pd.read_csv("BMSTransformations.csv")

"""

from __future__ import print_function

import sys
from os import walk
from os.path import join, dirname
from fnmatch import filter
import numpy as np
import h5py
from scri.SpEC import estimate_avg_com_motion as eacm


def run_in(filename, i_this, i_tot, f):
    print("Working on {0} -- {1} of {2}".format(filename, i_this, i_tot)); sys.stdout.flush()
    try:
        with h5py.File(filename, 'r') as horizons:
            tf = horizons['AhA.dir/ChristodoulouMass.dat'][-1, 0]
        vals = np.hstack(eacm(filename))
        dir, lev = dirname(filename).split('Lev')
        csv = '{0}, {1}, {2}, {3}\n'.format(dir, float(lev), ', '.join([str(f) for f in np.hstack(vals)]), tf)
        f.write(csv)
    except:
        print('Failed in {0} -- {1} of {2}'.format(filename, i_this, i_tot)); sys.stdout.flush()
    print('Finished {0} -- {1} of {2}'.format(filename, i_this, i_tot)); sys.stdout.flush()

if __name__ == '__main__':
    print("Finding files to operate on")
    files = [join(root, 'Horizons.h5')
             for top_dir in ['Catalog', 'Incoming']
             for root, dirnames, filenames in walk(top_dir)
             for filename in filter(filenames, 'Horizons.h5')]
    print("Finished finding {0} files to operate on".format(len(files)))
    with open("BMSTransformations.csv", 'w') as f:
        f.write('dirname,Lev,x0,y0,z0,vx0,vy0,vz0,t0,tf\n')
        for i_this, filename in enumerate(files, 1):
            run_in(filename, i_this, len(files), f)



# Copyright (c) 2015, Michael Boyle
# See LICENSE file for details: <https://github.com/moble/spherical_functions/blob/master/LICENSE>

from __future__ import print_function, division, absolute_import

import warnings
import numpy as np
from quaternion.numba_wrapper import jit, xrange
import spherical_functions as sf
from .. import (WaveformModes, FrameNames, DataType, DataNames, UnknownDataType, h, hdot, psi4)
from sxs.metadata import Metadata

def translate_data_types_GWFrames_to_waveforms(d):
    if d < 4:
        return {0: UnknownDataType, 1: h, 2: hdot, 3: psi4}[d]
    else:
        return DataType[d-4]


def translate_data_types_waveforms_to_GWFrames(d):
    if d in [UnknownDataType, h, hdot, psi4]:
        return {UnknownDataType: 0, h: 1, hdot: 2, psi4: 3}[d]
    else:
        return d+4


@jit
def index_is_monotonic(y):
    length = y.size
    monotonic = np.ones_like(y, dtype=bool)
    direction = y[-1] - y[0]
    if direction > 0.0:
        max_value = y[0]
        for i in xrange(1, length):
            if y[i] <= max_value:
                monotonic[i] = False
            else:
                max_value = y[i]
    else:
        min_value = y[0]
        for i in xrange(1, length):
            if y[i] >= min_value:
                monotonic[i] = False
            else:
                min_value = y[i]
    return monotonic


def monotonic_indices(y):
    indices = np.arange(y.size)
    return indices[index_is_monotonic(y)]


def monotonize(y):
    return y[monotonic_indices(y)]


def read_from_h5(file_name, **kwargs):
    """Read data from an H5 file in SXS format

    Note that SXS format is essentially compatible with NRAR format.  The existence of this function is not to be
    taken as the author's endorsement of either SXS or NRAR format.

    Parameters
    ----------
    file_name : str
        Path to H5 file containing the data, optionally including the path within the file itself to the directory
        containing the data.  For example, the standard SXS data with N=2 might be obtained with the file name
        `'rhOverM_Asymptotic_GeometricUnits.h5/Extrapolated_N2.dir'`.

    Keyword parameters
    ------------------
    frameType : int, optional
    dataType : int, optional
    r_is_scaled_out : bool, optional
    m_is_scaled_out : bool, optional
        These four parameters are documented in the docstring of the WaveformBase object.  Note that if any of these
        is present in the H5 file (which is not common) that value will override this argument.  If neither the file
        nor these parameters are present, defaults will be applied, assuming that the frame is inertial, R and M are
        both scaled out, and the data type (psi4, hdot, or h) can be gleaned from `file_name`.

    """

    import os.path
    import re
    import h5py
    import quaternion

    # This unfortunate concoction is needed to determine the (ell,m) values of the various mode data sets
    pattern_Ylm = re.compile(r"""Y_l(?P<L>[0-9]+)_m(?P<M>[-+0-9]+)\.dat""")

    # Initialize an empty object to be filled with goodies
    w = WaveformModes(constructor_statement='scri.SpEC.read_from_h5("{0}", **{1})'.format(file_name, kwargs))

    # Get an h5py handle to the desired part of the h5 file
    try:
        file_name, root_group = file_name.rsplit('.h5', 1)
        file_name += '.h5'
    except ValueError:
        root_group = ''  # FileName is just a file, not a group in a file
    try:
        f_h5 = h5py.File(file_name, 'r')
    except IOError:
        print("\n`read_from_h5` could not open the file '{0}'\n\n".format(file_name))
        raise
    if root_group:
        f = f_h5[root_group]
    else:
        f = f_h5

    # If it exists, add the metadata file to `w` as an object.  So, for example, the initial spin on
    # object 1 can be accessed as `w.metadata.initial_spin1`.  See the documentation of
    # `sxs.metadata.Metadata` for more details.  And in IPython, tab completion works on the
    # `w.metadata` object.
    try:
        w.metadata = Metadata.from_file(os.path.join(os.path.dirname(file_name), 'metadata'),
                                        ignore_invalid_lines=True, cache_json=False)
    except:
        pass  # Probably couldn't find the metadata.json/metadata.txt file

    try:  # Make sure the h5py.File gets closed, even in the event of an exception

        # Add the old history to the new history, if found
        try:
            try:
                old_history = f['History.txt'][()].decode()
            except AttributeError:
                old_history = f['History.txt'][()]
            w._append_history("", 0)
            w._append_history("<previous_history>", 2)
            w._append_history(old_history, 1)
            w._append_history("</previous_history>", 2)
        except KeyError:
            pass  # Did not find a history

        # Get the frame data, converting to quaternion objects
        try:
            w.frame = quaternion.as_quat_array(f['Frame'])
        except KeyError:
            pass  # There was no frame data

        # Get the descriptive items
        try:
            w.frameType = int(f.attrs['FrameType'])
        except KeyError:
            if 'frameType' in kwargs:
                w.frameType = int(kwargs.pop('frameType'))
            else:
                warning = ("\n`frameType` was not found in '{0}' or the keyword arguments.\n".format(file_name) +
                           "Using default value `{0}`.  You may want to set it manually.\n\n".format(FrameNames[1]))
                warnings.warn(warning)
                w.frameType = 1
        try:
            w.dataType = translate_data_types_GWFrames_to_waveforms(int(f.attrs['DataType']))
        except KeyError:
            if 'dataType' in kwargs:
                w.dataType = int(kwargs.pop('dataType'))
            else:
                found = False
                for type_int, type_name in zip(reversed(DataType), reversed(DataNames)):
                    if type_name.lower() in file_name.lower():
                        found = True
                        w.dataType = type_int
                        warning = ("\n`dataType` was not found in '{0}' or the keyword arguments.\n".format(file_name) +
                                   "Using default value `{0}`.  You may want to set it manually.".format(type_name))
                        warnings.warn(warning)
                        break
                if not found:
                    warning = ("\n`dataType` was not found in '{0}' or the keyword arguments.\n".format(file_name) +
                               "You may want to set it manually.")
                    warnings.warn(warning)
        try:
            w.r_is_scaled_out = bool(f.attrs['RIsScaledOut'])
        except KeyError:
            if 'r_is_scaled_out' in kwargs:
                w.r_is_scaled_out = bool(kwargs.pop('r_is_scaled_out'))
            else:
                warning = ("\n`r_is_scaled_out` was not found in '{0}' or the keyword arguments.\n".format(file_name) +
                           "Using default value `True`.  You may want to set it manually.\n\n")
                warnings.warn(warning)
                w.r_is_scaled_out = True
        try:
            w.m_is_scaled_out = bool(f.attrs['MIsScaledOut'])
        except KeyError:
            if 'm_is_scaled_out' in kwargs:
                w.m_is_scaled_out = bool(kwargs.pop('m_is_scaled_out'))
            else:
                warning = ("\n`m_is_scaled_out` was not found in '{0}' or the keyword arguments.\n".format(file_name) +
                           "Using default value `True`.  You may want to set it manually.\n\n")
                warnings.warn(warning)
                w.m_is_scaled_out = True

        # Get the names of all the data sets in the h5 file, and check for matches
        YLMdata = [data_set for data_set in list(f) for m in [pattern_Ylm.search(data_set)] if m]
        if len(YLMdata) == 0:
            raise ValueError("Couldn't understand data set names in '{0}'.\n".format(file_name) +
                             "Maybe you need to add the directory within the h5 file.\n" +
                             "E.g.: '{0}/Extrapolated_N2.dir'.".format(file_name))

        # Sort the data set names by increasing ell, then increasing m
        YLMdata = sorted(YLMdata, key=lambda data_set: [int(pattern_Ylm.search(data_set).group('L')),
                                                        int(pattern_Ylm.search(data_set).group('M'))])
        LM = np.array(sorted([[int(m.group('L')), int(m.group('M'))]
                              for data_set in YLMdata for m in [pattern_Ylm.search(data_set)] if m]))
        ell_min, ell_max = min(LM[:, 0]), max(LM[:, 0])
        if not np.array_equal(LM, sf.LM_range(ell_min, ell_max)):
            raise ValueError("Input [ell,m] modes are not complete.  Found modes:\n{0}\n".format(LM))
        n_modes = len(LM)

        # Get the time data (assuming all are equal)
        T = f[YLMdata[0]][:, 0]
        monotonic = index_is_monotonic(T)
        w.t = T[monotonic]
        n_times = len(w.t)

        # Loop through, setting data in each mode
        w.data = np.empty((n_times, n_modes), dtype=complex)
        for m, DataSet in enumerate(YLMdata):
            if f[DataSet].shape[0] != n_times:
                raise ValueError("The number of time steps in this dataset should be {0}; ".format(n_times) +
                                 "it is {0} in '{1}'.".format(f[DataSet].shape[0], DataSet))
            w.data[:, m] = f[DataSet][:, 1:3].view(dtype=np.complex)[monotonic, 0]

        # Now that the data is set, we can set these
        w.ells = ell_min, ell_max

        # Check up on the validity of the waveform
        if not w.ensure_validity(alter=True, assertions=False):
            raise ValueError("The data resulting from this input is invalid")

    except KeyError:
        print("\nThis H5 file appears to have not stored all the required information.\n\n")
        raise  # Re-raise the exception after adding our information

    finally:  # Use `finally` to make sure this happens:
        f_h5.close()

    if kwargs:
        import pprint
        warnings.warn("\nUnused kwargs passed to this function:\n{0}".format(pprint.pformat(kwargs, width=1)))

    return w


def write_to_h5(w, file_name, file_write_mode='w', attributes={}):
    """
    Output the Waveform in NRAR format.

    Note that the file_name is prepended with some descriptive information involving the data type and the frame type,
    such as 'rhOverM_Corotating_' or 'rMpsi4_Aligned_'.

    """

    import os.path
    import h5py
    import warnings

    group = None
    if '.h5' in file_name and not file_name.endswith('.h5'):
        file_name, group = file_name.split('.h5')
        file_name += '.h5'
    # Add descriptive prefix to file_name
    base_name = w.descriptor_string + '_' + os.path.basename(file_name)
    if not os.path.dirname(file_name):
        file_name = base_name
    else:
        file_name = os.path.join(os.path.dirname(file_name), base_name)
    # Open the file for output
    try:
        f = h5py.File(file_name, file_write_mode)
    except IOError:  # If that did not work...
        print("write_to_h5 was unable to open the file '{0}'.\n\n".format(file_name))
        raise  # re-raise the exception after the informative message above
    try:
        # If we are writing to a group within the file, create it
        if group:
            g = f.create_group(group)
        else:
            g = f
        # Now write all the data to various groups in the file
        g.attrs['OutputFormatVersion'] = 'scri.SpEC'
        g.create_dataset("History.txt", data='\n'.join(w.history) + '\n\nwrite_to_h5({0}, {1})\n'.format(w, file_name))
        g.attrs['FrameType'] = w.frameType
        g.attrs['DataType'] = translate_data_types_waveforms_to_GWFrames(w.dataType)
        g.attrs['RIsScaledOut'] = int(w.r_is_scaled_out)
        g.attrs['MIsScaledOut'] = int(w.m_is_scaled_out)
        for attr in attributes:
            try:
                g.attrs[attr] = attributes[attr]
            except:
                warning = "scri.SpEC.write_to_h5 unable to output attribute {0}={1}".format(attr, attributes[attr])
                warnings.warn(warning)
        for i_m in range(w.n_modes):
            ell, m = w.LM[i_m]
            Data_m = g.create_dataset("Y_l{0}_m{1}.dat".format(ell, m),
                                      data=[[t, d.real, d.imag] for t, d in zip(w.t, w.data[:, i_m])],
                                      compression="gzip", shuffle=True)
            Data_m.attrs['ell'] = ell
            Data_m.attrs['m'] = m
    finally:  # Use `finally` to make sure this happens:
        f.close()
